"""This module sets up a fake OIDC provider and saas fine-grained authorization endpoints
using FastAPI. It can be used for end-to-end tests without requiring an external identity provider.

The module provides:
    - All endpoints required by the OIDC standard.
    - Tenant mapping for authorization.

Note: the signed-in user is hardcoded and cannot be changed.
"""

import argparse
import base64
import html
import logging
import secrets
from binascii import unhexlify
from collections.abc import Mapping, Sequence
from typing import Annotated, Literal
from urllib.parse import urlencode

import fastapi as fapi
import jwt
import uvicorn
from cryptography.hazmat.primitives.asymmetric import rsa
from pydantic import BaseModel

from cmk_dev_site.cmk_dev_site import ensure_sudo
from cmk_dev_site.saas.config import AdminPanelUrlConfig, OIDCConfig
from cmk_dev_site.saas.constants import (
    ADMIN_PANEL_CONFIG_PATH,
    HOST,
    OIDC_CONFIG_PATH,
    OIDC_PORT,
    TENANT_ID,
    URL,
)
from cmk_dev_site.utils import is_port_in_use, write_root_owned_file
from cmk_dev_site.utils.cli import clean_cli_exit
from cmk_dev_site.utils.log import get_logger

application = fapi.FastAPI()
logger = get_logger(__name__)


class TenantInfo(BaseModel):
    user_role: Literal["user", "admin"]


class UserRoleAnswer(BaseModel):
    tenants: Mapping[str, TenantInfo]


class WellKnownResponseModel(BaseModel):
    authorization_endpoint: str
    token_endpoint: str
    jwks_uri: str
    issuer: str = "checkmk"
    scopes_supported: Sequence[str] = ["openid", "email"]
    response_types_supported: Sequence[str] = ["code", "token"]
    id_token_signing_alg_values_supported: Sequence[str] = ["RS256"]
    subject_types_supported: Sequence[str] = ["public"]
    token_endpoint_auth_methods_supported: Sequence[str] = ["client_secret_post"]
    grant_types_supported: Sequence[str] = ["authorization_code"]


class JWKS:
    def __init__(self) -> None:
        # this is a mock testsetup to key-size is not to important
        self.private = rsa.generate_private_key(public_exponent=65537, key_size=1024)
        self.public = self.private.public_key()
        self.kid = "usethis"

    @property
    def n(self) -> str:
        n = self.public.public_numbers().n
        hexi = hex(n).lstrip("0x")
        encoded = base64.urlsafe_b64encode(unhexlify(hexi))
        return encoded.decode("utf-8").rstrip("=")


KEY = JWKS()


class KeyModel(BaseModel):
    n: str
    alg: str = "RS256"
    e: str = "AQAB"
    kid: str
    use: str = "sig"
    kty: str = "RSA"


class JWKSModel(BaseModel):
    keys: Sequence[KeyModel]


class TokenResponse(BaseModel):
    id_token: str
    access_token: str


class TokenPayload(BaseModel):
    email: str
    aud: str
    sub: str = "1234567"
    user_role: Literal["user", "admin"]
    tenant_id: str = TENANT_ID


class AuthorizationCodeData(BaseModel):
    username: str
    user_role: Literal["user", "admin"]


DEFAULT_USERNAME = "test@test.com"
DEFAULT_ROLE: Literal["user", "admin"] = "admin"
AUTHORIZATION_CODES: dict[str, AuthorizationCodeData] = {}


@application.get("/.well-known/openid-configuration", status_code=200)
def well_known() -> WellKnownResponseModel:
    return WellKnownResponseModel(
        authorization_endpoint=f"{URL}/authorize",
        jwks_uri=f"{URL}/.well-known/jwks.json",
        token_endpoint=f"{URL}/token",
    )


@application.get("/.well-known/jwks.json", response_model=JWKSModel)
def jwks() -> JWKSModel:
    key = KeyModel(n=KEY.n, kid=KEY.kid)
    return JWKSModel(keys=[key])


@application.get("/healthz", status_code=200, responses={200: {}})
def liveness() -> str:
    return "I'm alive"


@application.post("/token", response_model=TokenResponse)
def token(
    client_id: Annotated[str, fapi.Form()], code: Annotated[str, fapi.Form()]
) -> TokenResponse:
    try:
        login_data = AUTHORIZATION_CODES.pop(code)
    except KeyError as exc:
        raise fapi.HTTPException(status_code=400, detail="Invalid authorization code") from exc

    payload = TokenPayload(
        email=login_data.username,
        aud=client_id,
        sub=login_data.username,
        user_role=login_data.user_role,
    )
    id_token = jwt.encode(
        payload.model_dump(), KEY.private, algorithm="RS256", headers={"kid": KEY.kid}
    )
    return TokenResponse(id_token=id_token, access_token=id_token)


@application.get("/api/users/me/tenants")
def tenant_role_mapping(
    authorization: Annotated[str | None, fapi.Header()] = None,
) -> UserRoleAnswer:
    if authorization is None:
        raise fapi.HTTPException(status_code=401, detail="Missing Authorization header")

    scheme, _, token_str = authorization.partition(" ")
    if scheme.lower() != "bearer" or not token_str:
        raise fapi.HTTPException(status_code=401, detail="Invalid Authorization header")

    try:
        decoded = jwt.decode(
            token_str, KEY.public, algorithms=["RS256"], options={"verify_aud": False}
        )
        payload = TokenPayload.model_validate(decoded)
    except jwt.InvalidTokenError as exc:
        raise fapi.HTTPException(status_code=401, detail="Invalid token") from exc

    return UserRoleAnswer(tenants={payload.tenant_id: TenantInfo(user_role=payload.user_role)})


@application.get("/logout")
def logout(client_id: str, redirect_uri: str) -> fapi.responses.RedirectResponse:
    return fapi.responses.RedirectResponse(redirect_uri)


def _parse_args() -> argparse.Namespace:
    p = argparse.ArgumentParser()
    p.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="increase output verbosity",
    )
    return p.parse_args()


def run() -> None:
    with clean_cli_exit():
        args = _parse_args()
        log_level = logging.DEBUG if args.verbose >= 1 else logging.INFO
        logger.setLevel(log_level)

        if is_port_in_use(OIDC_PORT):
            logger.info("OIDC port is used. Assume fake provider is running")
            return

        ensure_sudo()

        logger.debug("writing config")
        config = OIDCConfig()
        write_root_owned_file(OIDC_CONFIG_PATH, config.model_dump_json(indent=4))

        admin_panel = AdminPanelUrlConfig()
        write_root_owned_file(ADMIN_PANEL_CONFIG_PATH, admin_panel.model_dump_json(indent=4))

        logger.debug("starting uvicorn")
        uvicorn.run(application, port=OIDC_PORT, host=HOST, log_level=log_level)


@application.get("/authorize")
def authorize(state: str, redirect_uri: str) -> fapi.responses.HTMLResponse:
    escaped_state = html.escape(state, quote=True)
    escaped_redirect_uri = html.escape(redirect_uri, quote=True)
    escaped_username = html.escape(DEFAULT_USERNAME, quote=True)
    checked_user = " checked" if DEFAULT_ROLE == "user" else ""
    checked_admin = " checked" if DEFAULT_ROLE == "admin" else ""

    page = f"""<!doctype html>
<html>
<body>
<form method="post" action="/authorize">
<input type="hidden" name="state" value="{escaped_state}">
<input type="hidden" name="redirect_uri" value="{escaped_redirect_uri}">
<label>Username <input type="text" name="username" value="{escaped_username}"></label>
<label><input type="radio" name="role" value="user"{checked_user}> user</label>
<label><input type="radio" name="role" value="admin"{checked_admin}> admin</label>
<button type="submit">Login</button>
</form>
</body>
</html>"""
    return fapi.responses.HTMLResponse(page)


@application.post("/authorize")
def authorize_login(
    state: Annotated[str, fapi.Form()],
    redirect_uri: Annotated[str, fapi.Form()],
    username: Annotated[str, fapi.Form()],
    role: Annotated[Literal["user", "admin"], fapi.Form()],
) -> fapi.responses.RedirectResponse:
    code = secrets.token_urlsafe(16)
    AUTHORIZATION_CODES[code] = AuthorizationCodeData(username=username, user_role=role)
    params = {"state": state, "code": code}
    url = f"{redirect_uri}?{urlencode(params)}"
    return fapi.responses.RedirectResponse(url, status_code=303)



import difflib
import functools
import subprocess
from dataclasses import dataclass
from datetime import date
from enum import StrEnum
from typing import Any, Literal

from typing_extensions import override


class Edition(StrEnum):
    # old
    OLD_RAW = "cre"
    OLD_ENTERPRISE = "cee"
    OLD_MANAGED = "cme"
    OLD_CLOUD = "cce"
    OLD_SAAS = "cse"
    # new
    COMMUNITY = "community"
    PRO = "pro"
    ULTIMATEMT = "ultimatemt"
    ULTIMATE = "ultimate"
    CLOUD = "cloud"


def _map_edition_to_package_name(edition: Edition) -> str:
    match edition:
        case Edition.OLD_RAW:
            return "raw"
        case Edition.OLD_ENTERPRISE:
            return "enterprise"
        case Edition.OLD_MANAGED:
            return "managed"
        case Edition.OLD_CLOUD:
            return "cloud"
        case Edition.OLD_SAAS:
            return "saas"
        case _:
            return edition.value


@functools.total_ordering
class BaseVersion:
    def __init__(self, major: int, minor: int, patch: int = 0):
        self.major = major
        self.minor = minor
        self.patch = patch

    @classmethod
    def from_str(cls, version_str: str) -> "BaseVersion":
        """Create a Version object from a string.
        This method expects the version string to be in the format 'd.d[.d]'.
        """
        match version_str.split("."):
            case (major_str, minor_str):
                return cls(major=int(major_str), minor=int(minor_str))
            case (major_str, minor_str, patch_str):
                return cls(major=int(major_str), minor=int(minor_str), patch=int(patch_str))
            case _:
                raise ValueError("Version must be in 'd.d[.d]' format")

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"

    def __eq__(self, other: Any):
        if isinstance(other, BaseVersion):
            return (self.major, self.minor, self.patch) == (
                other.major,
                other.minor,
                other.patch,
            )
        return NotImplemented

    def __lt__(self, other: Any):
        if isinstance(other, BaseVersion):
            return (self.major, self.minor, self.patch) < (
                other.major,
                other.minor,
                other.patch,
            )

    def iso_format(self) -> str:
        """Return the version in ISO format."""
        return self.__str__()

    @property
    def download_folder_name(self) -> str:
        """Get the download folder name for this version."""
        return self.__str__()


class VersionWithPatch(BaseVersion):
    def __init__(self, base_version: BaseVersion, patch_type: Literal["p", "b"], patch: int):
        self.base_version = base_version
        self.patch_type = patch_type
        self.patch = patch

    def __str__(self) -> str:
        if self.patch_type == "p" and self.patch == 0:
            return f"{self.base_version}"
        return f"{self.base_version}{self.patch_type}{self.patch}"

    def iso_format(self) -> str:
        return self.__str__()


class VersionWithReleaseCandidate(BaseVersion):
    def __init__(
        self, base_version: BaseVersion, patch_type: Literal["p", "b"], patch: int, rc: int
    ):
        self.base_version = base_version
        self.patch_type = patch_type
        self.patch = patch
        self.rc = rc

    def __str__(self) -> str:
        return f"{self.base_version}{self.patch_type}{self.patch}"

    def iso_format(self) -> str:
        return self.__str__()

    @property
    @override
    def download_folder_name(self) -> str:
        return f"{self.base_version}{self.patch_type}{self.patch}-rc{self.rc}"


class VersionWithReleaseDate(BaseVersion):
    def __init__(self, base_version: BaseVersion, release_date: date):
        self.base_version = base_version
        self.release_date = release_date

    def __str__(self) -> str:
        return f"{self.base_version}-{self.release_date.strftime('%Y.%m.%d')}"

    def __repr__(self) -> str:
        return self.__str__()

    def iso_format(self) -> str:
        return f"{self.base_version}-{self.release_date.isoformat()}"


class PartialVersion(BaseVersion):
    """Represents a version that release date is not known."""

    pass


class GitVersion:
    def __init__(self, branch: str, commit_hash: str):
        self.branch = branch
        self.commit_hash = commit_hash

    def __str__(self) -> str:
        return f"git:{self.branch}:{self.commit_hash}"


Version = (
    BaseVersion
    | VersionWithPatch
    | VersionWithReleaseDate
    | PartialVersion
    | GitVersion
    | VersionWithReleaseCandidate
)


class CMKPackage:
    """Represents a Checkmk package."""

    def __init__(
        self,
        version: VersionWithPatch
        | VersionWithReleaseDate
        | BaseVersion
        | VersionWithReleaseCandidate,
        edition: Edition,
        distro_codename: str = "noble",
        arch: str = "amd64",
    ):
        self.version = version
        self.edition = edition
        self.distro_codename = distro_codename
        self.arch = arch

    @property
    def omd_version(self) -> str:
        """Get the OMD version string."""
        return f"{self.version}.{self.edition.value}"

    @property
    def package_raw_name(self) -> str:
        """Get the package name."""
        base = f"check-mk-{_map_edition_to_package_name(self.edition)}-"
        if isinstance(self.version, VersionWithReleaseCandidate):
            return f"{base}{self.version.base_version}{self.version.patch_type}{self.version.patch}"
        else:
            return f"{base}{self.version}"

    @property
    def package_name(self) -> str:
        """Get the package name."""
        return f"{self.package_raw_name}_0.{self.distro_codename}_{self.arch}.deb"

    @property
    def base_version(self) -> BaseVersion:
        """Get the base version."""
        match self.version:
            case VersionWithPatch(base_version=base_version, patch_type=_, patch=_):
                return base_version
            case VersionWithReleaseDate(base_version=base_version, release_date=_):
                return base_version
            case VersionWithReleaseCandidate(
                base_version=base_version, patch_type=_, patch=_, rc=_
            ):
                return base_version
            case BaseVersion():
                return self.version

    def __str__(self) -> str:
        return f"{self.version}.{self.edition.value}"

    def __repr__(self) -> str:
        return self.__str__()


@dataclass
class PartialCMKPackage:
    version: str

    def similarity(self, other: str) -> float:
        return difflib.SequenceMatcher(None, self.version, other).ratio()


def omd_config_set(site_name: str, config_key: str, config_value: str) -> None:
    try:
        result = subprocess.run(
            [
                "sudo",
                "omd",
                "config",
                site_name,
                "set",
                config_key,
                config_value,
            ],
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            f"Could not set configuration {config_key} to {config_value} "
            f"for site {site_name}\n{e.stderr}"
        ) from e

    if result.stderr:
        raise RuntimeError(result.stderr.decode("utf-8").strip())


def omd_config_get(site_name: str, param: str) -> str | None:
    try:
        return subprocess.run(
            [
                "sudo",
                "su",
                site_name,
                "-c",
                f"omd config show {param}",
            ],
            check=True,
            capture_output=True,
            text=True,
        ).stdout.split("\n")[0]

    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            f"Could not get configuration {param} for site {site_name}\n{e.stderr}"
        ) from e



from datetime import date

import pytest

from cmk_dev_site.cmk_dev_install import build_download_url
from cmk_dev_site.omd import (
    BaseVersion,
    CMKPackage,
    Edition,
    VersionWithPatch,
    VersionWithReleaseCandidate,
    VersionWithReleaseDate,
)


def test_base_version_from_str() -> None:
    assert BaseVersion.from_str("2.3") == BaseVersion(2, 3)
    assert BaseVersion.from_str("2.3.0") == BaseVersion(2, 3, 0)
    with pytest.raises(ValueError):
        BaseVersion.from_str("2")
    with pytest.raises(ValueError):
        BaseVersion.from_str("2.3.0.1")


def test_base_version_str() -> None:
    assert str(BaseVersion(2, 3)) == "2.3.0"
    assert str(BaseVersion(2, 3, 1)) == "2.3.1"


def test_base_version_eq() -> None:
    assert BaseVersion(2, 3) == BaseVersion(2, 3, 0)
    assert BaseVersion(2, 3) != BaseVersion(2, 3, 1)


def test_base_version_lt() -> None:
    assert BaseVersion(2, 3) < BaseVersion(2, 3, 1)
    assert BaseVersion(2, 3) < BaseVersion(2, 4)
    assert BaseVersion(2, 3) < BaseVersion(3, 0)


def test_version_with_patch_str() -> None:
    assert str(VersionWithPatch(BaseVersion(2, 3), "p", 1)) == "2.3.0p1"


def test_version_with_release_candidate_str() -> None:
    assert str(VersionWithReleaseCandidate(BaseVersion(2, 3), "p", 1, 1)) == "2.3.0p1"


def test_version_with_release_candidate_download_folder_name() -> None:
    assert (
        VersionWithReleaseCandidate(BaseVersion(2, 3), "p", 1, 1).download_folder_name
        == "2.3.0p1-rc1"
    )


def test_version_with_release_date_str() -> None:
    assert str(VersionWithReleaseDate(BaseVersion(2, 3), date(2025, 1, 1))) == "2.3.0-2025.01.01"


def test_version_with_release_date_iso_format() -> None:
    assert (
        VersionWithReleaseDate(BaseVersion(2, 3), date(2025, 1, 1)).iso_format()
        == "2.3.0-2025-01-01"
    )


def test_cmk_package_omd_version() -> None:
    # Test with BaseVersion
    pkg_base = CMKPackage(BaseVersion(2, 3), Edition.COMMUNITY)
    assert pkg_base.omd_version == "2.3.0.community"

    # Test with VersionWithPatch
    pkg_patch = CMKPackage(VersionWithPatch(BaseVersion(2, 3), "p", 1), Edition.PRO)
    assert pkg_patch.omd_version == "2.3.0p1.pro"

    # Test with VersionWithReleaseDate
    pkg_date = CMKPackage(
        VersionWithReleaseDate(BaseVersion(2, 3), date(2025, 1, 1)), Edition.ULTIMATE
    )
    assert pkg_date.omd_version == "2.3.0-2025.01.01.ultimate"

    # Test with VersionWithReleaseCandidate
    pkg_rc = CMKPackage(VersionWithReleaseCandidate(BaseVersion(2, 3), "p", 1, 1), Edition.CLOUD)
    assert pkg_rc.omd_version == "2.3.0p1.cloud"


def test_cmk_package_package_raw_name() -> None:
    # Test with BaseVersion
    pkg_base = CMKPackage(BaseVersion(2, 3), Edition.COMMUNITY)
    assert pkg_base.package_raw_name == "check-mk-community-2.3.0"

    # Test with VersionWithPatch
    pkg_patch = CMKPackage(VersionWithPatch(BaseVersion(2, 3), "p", 1), Edition.PRO)
    assert pkg_patch.package_raw_name == "check-mk-pro-2.3.0p1"

    # Test with VersionWithReleaseDate
    pkg_date = CMKPackage(
        VersionWithReleaseDate(BaseVersion(2, 3), date(2025, 1, 1)), Edition.ULTIMATE
    )
    assert pkg_date.package_raw_name == "check-mk-ultimate-2.3.0-2025.01.01"

    # Test with VersionWithReleaseCandidate
    pkg_rc = CMKPackage(VersionWithReleaseCandidate(BaseVersion(2, 3), "p", 1, 1), Edition.CLOUD)
    assert pkg_rc.package_raw_name == "check-mk-cloud-2.3.0p1"


def test_cmk_package_package_name() -> None:
    # Test with BaseVersion
    pkg_base = CMKPackage(BaseVersion(2, 3), Edition.COMMUNITY)
    assert pkg_base.package_name == "check-mk-community-2.3.0_0.noble_amd64.deb"

    # Test with VersionWithPatch
    pkg_patch = CMKPackage(
        VersionWithPatch(BaseVersion(2, 3), "p", 1),
        Edition.PRO,
        distro_codename="focal",
        arch="arm64",
    )
    assert pkg_patch.package_name == "check-mk-pro-2.3.0p1_0.focal_arm64.deb"

    # Test with VersionWithReleaseDate
    pkg_date = CMKPackage(
        VersionWithReleaseDate(BaseVersion(2, 3), date(2025, 1, 1)), Edition.ULTIMATE
    )
    assert pkg_date.package_name == "check-mk-ultimate-2.3.0-2025.01.01_0.noble_amd64.deb"

    # Test with VersionWithReleaseCandidate
    pkg_rc = CMKPackage(VersionWithReleaseCandidate(BaseVersion(2, 3), "p", 1, 1), Edition.CLOUD)
    assert pkg_rc.package_name == "check-mk-cloud-2.3.0p1_0.noble_amd64.deb"


def test_cmk_package_base_version() -> None:
    assert CMKPackage(BaseVersion(2, 3), Edition.COMMUNITY).base_version == BaseVersion(2, 3)
    assert CMKPackage(
        VersionWithPatch(BaseVersion(2, 3), "p", 1), Edition.COMMUNITY
    ).base_version == BaseVersion(2, 3)
    assert CMKPackage(
        VersionWithReleaseDate(BaseVersion(2, 3), date(2025, 1, 1)),
        Edition.COMMUNITY,
    ).base_version == BaseVersion(2, 3)
    assert CMKPackage(
        VersionWithReleaseCandidate(BaseVersion(2, 3), "p", 1, 1),
        Edition.COMMUNITY,
    ).base_version == BaseVersion(2, 3)


def test_build_download_url() -> None:
    base_url = "https://download.checkmk.com/checkmk"
    pkg = CMKPackage(BaseVersion(2, 3), Edition.COMMUNITY)
    expected_url = (
        "https://download.checkmk.com/checkmk/2.3.0/check-mk-community-2.3.0_0.noble_amd64.deb"
    )
    assert build_download_url(base_url, pkg) == expected_url

    pkg_patch = CMKPackage(VersionWithPatch(BaseVersion(2, 3), "p", 1), Edition.PRO)
    expected_url_patch = (
        "https://download.checkmk.com/checkmk/2.3.0p1/check-mk-pro-2.3.0p1_0.noble_amd64.deb"
    )
    assert build_download_url(base_url, pkg_patch) == expected_url_patch

    pkg_rc = CMKPackage(VersionWithReleaseCandidate(BaseVersion(2, 3), "p", 1, 1), Edition.CLOUD)
    expected_url_rc = (
        "https://download.checkmk.com/checkmk/2.3.0p1-rc1/check-mk-cloud-2.3.0p1_0.noble_amd64.deb"
    )
    assert build_download_url(base_url, pkg_rc) == expected_url_rc

    pkg_date = CMKPackage(
        VersionWithReleaseDate(BaseVersion(2, 3), date(2025, 1, 1)), Edition.ULTIMATE
    )
    expected_url_date = "https://download.checkmk.com/checkmk/2.3.0-2025.01.01/check-mk-ultimate-2.3.0-2025.01.01_0.noble_amd64.deb"
    assert build_download_url(base_url, pkg_date) == expected_url_date



import signal
import sys
from collections.abc import Generator
from contextlib import contextmanager
from types import FrameType


@contextmanager
def clean_cli_exit() -> Generator[None, None, None]:
    """Context manager for clean keyboard interrupt and signal handling."""

    def signal_handler(signum: int, frame: FrameType | None) -> None:
        print(f"\n\nReceived signal {signum}.", file=sys.stderr)
        sys.exit(128 + signum)

    for sig in [signal.SIGTERM, signal.SIGHUP]:
        signal.signal(sig, signal_handler)

    try:
        yield
    except KeyboardInterrupt:
        print("\n\nInterrupted by user (Ctrl-C).", file=sys.stderr)
        sys.exit(130)



import functools
import inspect
import logging
import sys
from collections.abc import Callable
from typing import Any, ClassVar, ParamSpec, TextIO, TypeVar, cast

PROGRESS_LEVEL = 25  # Between  INFO (20) and WARNING (30)


def colorize(text: str, color: str) -> str:
    """Colorize text with ANSI escape codes."""
    colors = {
        "blue": "\033[34m",
        "green": "\033[32m",
        "yellow": "\033[33m",
        "red": "\033[31m",
        "magenta": "\033[35;1m",
        "cyan": "\033[36m",
    }
    rest = "\033[0m"
    if color_code := colors.get(color):
        return f"{color_code}{text}{rest}"

    return text


P = ParamSpec("P")
T = TypeVar("T")


def generate_log_decorator(
    logger: logging.Logger,
):
    def log(
        info_message: str | None = None,
        error_message: str | None = None,
        message_info: Callable[P, str] | None = None,
        prefix: Callable[P, str] | None = None,
        max_level: int = logging.INFO,
        show_result: bool = False,
    ) -> Callable[[Callable[P, T]], Callable[P, T]]:
        """Decorator for logging function calls, handling errors with optional warnings

        :param info_message: The message to log before the function call.
        :param error_message: The message to log if the function raises an exception.
        :param message_info: A function to generate a custom info message based on the function's
            arguments.
        :param prefix: A function to generate a custom prefix for the info message based on the
            function's arguments.
        :param max_level: maximum logging level to use for logging message.
        """

        def decorator(func: Callable[P, T]) -> Callable[P, T]:
            @functools.wraps(func)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                formatted_func_name = func.__name__.replace("_", " ").capitalize()
                msg = (
                    message_info(*args, **kwargs)
                    if message_info
                    else info_message or f"{formatted_func_name}..."
                )

                msg = f"{prefix(*args, **kwargs) if prefix else ''}{msg}"

                # Log function call details
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(
                        "Calling %s with args: %s, kwargs: %s",
                        colorize(func.__name__, "cyan"),
                        args,
                        kwargs,
                    )
                elif max_level >= logging.INFO:
                    logger.log(PROGRESS_LEVEL, msg)

                try:
                    result = func(*args, **kwargs)

                except Exception:  # pylint: disable=broad-except
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("%s Trace:", func.__name__, exc_info=True)
                    else:
                        logger.info("%s%s", msg, colorize("Failed", "red"))
                    if error_message:
                        logger.error(error_message)

                    # re-raise the exception
                    raise
                else:
                    # Log function result details
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("%s returned: %s", colorize(func.__name__, "cyan"), result)
                    elif max_level >= logging.INFO:
                        if show_result:
                            logger.info("%s", f"{msg}{colorize(str(result), 'green')}")
                        else:
                            logger.info("%s", f"{msg}{colorize('OK', 'green')}")

                    return result

            return wrapper

        return decorator

    return log


class InlineStreamHandler(logging.StreamHandler[TextIO]):
    """Custom StreamHandler to support inline logging for progress updates."""

    def __init__(self):
        super().__init__(sys.stderr)
        self.current_inline_message = False

    def emit(self, record: logging.LogRecord) -> None:
        # Format the message
        msg = self.format(record)
        if record.levelno == PROGRESS_LEVEL:
            # Overwrite the same line for progress
            print(f"\r\033[K{msg}", end="", flush=True, file=sys.stderr)
            self.current_inline_message = True
        else:
            # Clear inline message if switching to a regular log
            if self.current_inline_message:
                print(f"\r\033[K{msg}", flush=True, file=sys.stderr)
                self.current_inline_message = False
            else:
                print(msg, flush=True, file=sys.stderr)


class ColoredFormatter(logging.Formatter):
    """Custom log formatter that colorizes log messages based on their level."""

    # Define ANSI escape codes for colors
    COLORS: ClassVar = {
        logging.DEBUG: "blue",  # Blue
        logging.INFO: "green",  # Green
        logging.WARNING: "yellow",  # Yellow
        logging.ERROR: "red",  # Red
        logging.CRITICAL: "magenta",  # Bright Magenta
        PROGRESS_LEVEL: "cyan",  # Cyan
    }

    def format(self, record: logging.LogRecord) -> str:
        color = self.COLORS.get(record.levelno, "")

        # Colorize the levelname using termcolor
        record.levelname = colorize(record.levelname, color)

        # Call the parent format method to handle the message formatting
        return super().format(record)


class CustomLogger(logging.Logger):
    def progress(self, msg: str, *args: Any, **kwargs: Any):
        if self.isEnabledFor(PROGRESS_LEVEL):
            self.log(PROGRESS_LEVEL, msg, *args, **kwargs)


def get_logger(name: str) -> CustomLogger:
    logging.addLevelName(PROGRESS_LEVEL, "INFO")

    logging.setLoggerClass(CustomLogger)
    logger: CustomLogger = cast(CustomLogger, logging.getLogger(name))
    logging.setLoggerClass(logging.Logger)

    # Use the ColoredFormatter
    formatter = ColoredFormatter("%(levelname)s: %(message)s")
    console_handler = InlineStreamHandler()
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    return logger


def add_method_logging(
    log: Callable[[Callable[P, T]], Callable[P, T]],
    methods: list[str] | None = None,
    prefix: str = "",
    exclude_methods: list[str] | None = None,
):
    """
    Class decorator that adds logging to specified methods or all public methods

    Args:
        methods: List of method names to log. If None, logs all public methods
        prefix: Prefix for log messages
        exclude_methods: Methods to exclude from logging
    """

    def class_decorator(cls: type[T]) -> type[T]:
        exclude_methods_set = set(exclude_methods or [])
        exclude_methods_set.update(["__init__", "__new__", "__del__"])

        if methods is None:
            # Get all public methods
            target_methods: list[Any] = [
                name
                for name, _method in inspect.getmembers(cls, predicate=inspect.isfunction)
                if not name.startswith("_") and name not in exclude_methods_set
            ]
        else:
            target_methods = [m for m in methods if m not in exclude_methods_set]

        for method_name in target_methods:
            if hasattr(cls, method_name):
                original_method = getattr(cls, method_name)
                decorated_method = log(original_method)
                setattr(cls, method_name, decorated_method)

        return cls

    return class_decorator



import logging
import socket
import subprocess
from collections.abc import Sequence
from pathlib import Path
from typing import Any

from .log import colorize


def run_command(
    args: Sequence[str],
    check: bool = True,
    error_message: str | None = None,
    raise_runtime_error: bool = True,
    text: bool = True,
    logger: logging.Logger | None = logging.getLogger(__name__),
    silent: bool = False,
    **kwargs: Any,
) -> subprocess.CompletedProcess[Any]:
    result = subprocess.run(args=args, capture_output=True, text=text, **kwargs)
    msg = f"Running command {colorize(' '.join(args), 'blue')}:"
    if not silent and result.stdout:
        msg = msg + "\nSTDOUT:\n" + colorize(result.stdout.strip(), "green")
    if not silent and result.stderr:
        msg = msg + "\nSTDERR:\n" + colorize(result.stderr.strip(), "red")
    if logger:
        logger.debug(msg)
    if check and result.returncode != 0:
        if raise_runtime_error:
            error_msg = (error_message + "\n") if error_message else ""
            error_msg += (
                f"ERROR: Command failed: {colorize(' '.join(args), 'red')}\n"
                f"stdout: {result.stdout}\n"
                f"stderr: {result.stderr}"
            )
            raise RuntimeError(error_msg)
        else:
            if logger and error_message:
                logger.warning(error_message)

    return result


def is_port_in_use(port: int, host: str = "127.0.0.1") -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind((host, port))
            return False
        except OSError:
            return True


def write_root_owned_file(path: Path, content: str, permissions: str = "0666") -> None:
    """Write content to a root-owned file with specified permissions.

    Creates the parent directory if needed, writes the content via sudo,
    and sets the specified permissions. Content is piped through sudo
    to ensure secure writing regardless of permission settings.

    Args:
        path: Path to the file to create/write
        content: String content to write to the file
        permissions: Octal permission string (default: "0666" for world-writable)
    """
    # Create parent directory if needed
    run_command(["sudo", "mkdir", "-p", str(path.parent)])

    # Write content via sudo using tee (silent output)
    run_command(
        ["sudo", "tee", str(path)],
        input=content,
        silent=True,
    )

    # Set permissions
    run_command(["sudo", "chmod", permissions, str(path)])



from pydantic import BaseModel

from cmk_dev_site.saas.constants import TENANT_ID, URL


class OIDCConfig(BaseModel):
    base_url: str = "http://localhost"
    client_id: str = "notused"
    tenant_id: str = TENANT_ID
    well_known: str = f"{URL}/.well-known/openid-configuration"
    saas_api_url: str = URL
    logout_url: str = f"{URL}/logout"


class AdminPanelUrlConfig(BaseModel):
    uap_url: str = "https://admin-panel.saas-prod.cloudsandbox.checkmk.cloud/"
    bug_tracker_url: str = "https://admin-panel.saas-prod.cloudsandbox.checkmk.cloud/bug-report"
    download_agent_user: str = "automation"
    tenant_id: str = TENANT_ID



from cmk_dev_site.utils.log import colorize


def test_colorize():
    assert colorize("test", "blue") == "\033[34mtest\033[0m"
    assert colorize("test", "red") == "\033[31mtest\033[0m"
    assert colorize("test", "unknown") == "test"



"""cmk-dev-site

Set up sites based on the specified OMD version
Defaults to the current OMD version if version is not provided.
If a site with the same name and version already exists, it skips recreation and base
configuration, proceeding with the next steps. Use the -f option to force a full setup.
"""

import argparse
import getpass
import logging
import re
import shutil
import subprocess
import sys
import time
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime
from enum import StrEnum
from pathlib import Path
from typing import (
    Any,
    Self,
)

from cmk_dev_site.saas.constants import OIDC_PORT

from .cmk.rest_api import APIClient, CheckmkAPIException, RemoteSiteConnectionConfig
from .omd import (
    BaseVersion,
    CMKPackage,
    Edition,
    PartialCMKPackage,
    VersionWithPatch,
    VersionWithReleaseDate,
    omd_config_get,
    omd_config_set,
)
from .utils import is_port_in_use
from .utils import run_command as run_command_c
from .utils.cli import clean_cli_exit
from .utils.log import add_method_logging, colorize, generate_log_decorator, get_logger
from .version import __version__

logger = get_logger(__name__)
log = generate_log_decorator(logger)
omd_config_set = log(max_level=logging.DEBUG)(omd_config_set)
omd_config_get = log(max_level=logging.DEBUG)(omd_config_get)


def run_command(
    args: Sequence[str],
    check: bool = True,
    error_message: str | None = None,
    raise_runtime_error: bool = True,
    text: bool = True,
    silent: bool = False,
    **kwargs: Any,
) -> subprocess.CompletedProcess[Any]:
    return run_command_c(
        args,
        check=check,
        error_message=error_message,
        raise_runtime_error=raise_runtime_error,
        text=text,
        silent=silent,
        logger=logger,  # slightly hackish: pass in the right logger
        **kwargs,
    )


def _prefix_log_api_client(self: APIClient, *_args: Any, **_kwargs: Any) -> str:
    return f"[{colorize(self.site_name, 'blue')}]: "


add_method_logging(log(prefix=_prefix_log_api_client))(APIClient)

GUI_USER = "cmkadmin"
GUI_PW = "cmk"
INSTALLATION_PATH = Path("/omd/versions")


class Language(StrEnum):
    EN = "en"
    DE = "de"
    RO = "ro"


def _prefix_log_site(self: "Site", *args: Any, **kwargs: Any) -> str:
    return f"[{colorize(self.name, 'blue')}]: "


class Site:
    def __init__(
        self,
        site_name: str,
        cmk_pkg: CMKPackage,
        *,
        is_remote: bool = False,
    ):
        self.name = site_name
        self.cmk_pkg = cmk_pkg
        self._is_remote = is_remote

    def __repr__(self):
        return (
            f"<Site name={self.name} cmk_pkg={self.cmk_pkg} is_remote_site={self.is_remote_site}>"
        )

    @property
    def is_remote_site(self):
        return self._is_remote

    @log(prefix=_prefix_log_site)
    def create_site(self) -> None:
        run_command(
            [
                "sudo",
                "omd",
                "-V",
                self.cmk_pkg.omd_version,
                "create",
                "--apache-reload",
                "--no-autostart",
                self.name,
            ],
            check=True,
            error_message=f"[{self.name}]: Failed to create the site",
        )

        # Setting the password for user
        hashalgo = "-m" if self.cmk_pkg.base_version == "2.1.0" else "-B"

        run_command(
            [
                "sudo",
                "htpasswd",
                "-b",
                hashalgo,
                f"/omd/sites/{self.name}/etc/htpasswd",
                GUI_USER,
                GUI_PW,
            ],
            check=True,
            error_message=f"[{self.name}]: Failed to create site user pass",
        )

    def ensure_auth_works(self) -> None:
        if self.cmk_pkg.edition != Edition.OLD_SAAS and self.cmk_pkg.edition != Edition.CLOUD:
            return
        if not is_port_in_use(OIDC_PORT):
            raise RuntimeError(
                """Expect mock auth to be running for cloud site.
                Use `cmk-dev-site-mock-auth` command to run it"""
            )

    @log(prefix=_prefix_log_site)
    def delete_site(self) -> None:
        """Delete the site if it already exists."""
        if Path("/omd/sites", self.name).exists():
            run_command(
                [
                    "sudo",
                    "omd",
                    "-f",
                    "rm",
                    "--kill",
                    "--apache-reload",
                    self.name,
                ],
                check=True,
            )

    @log(prefix=_prefix_log_site)
    def configure_site(self, configs: list[tuple[str, str]]) -> None:
        try:
            if self.cmk_pkg.base_version == "2.1.0" and "MKEVENTD" not in dict(configs):
                omd_config_set(self.name, "MKEVENTD", "on")

            if "LIVESTATUS_TCP" not in dict(configs):
                omd_config_set(self.name, "LIVESTATUS_TCP", "on")

        except RuntimeError as e:
            logger.warning(
                "[%s]: Failed to configure the site. %s",
                colorize(self.name, "yellow"),
                e,
            )

        for key, value in configs:
            try:
                omd_config_set(self.name, key, value)
            except RuntimeError as e:
                logger.warning(
                    "[%s]: Failed to set option %s=%s. %s",
                    colorize(self.name, "yellow"),
                    key,
                    value,
                    e,
                )

    @log(prefix=_prefix_log_site)
    def start_site(self, api: "APIClient") -> None:
        run_command(
            ["sudo", "omd", "start", self.name],
            check=True,
            raise_runtime_error=False,
            error_message=(
                f"[{colorize(self.name, 'yellow')}]: Failed to start the site. "
                "Site probably is running"
            ),
        )
        max_try = 5
        while api.version() is None and max_try > 0:
            max_try -= 1
            time.sleep(1)
        if max_try == 0:
            raise RuntimeError(
                f"[{colorize(self.name, 'red')}]: Failed to start the site. "
                "Check if the site is running and try again."
            )

    @log(prefix=_prefix_log_site)
    def trigger_site_checking_cycle(self) -> None:
        run_command(
            ["sudo", "su", "-", self.name, "-c", f"cmk -n '{self.name}'"],
            error_message=(
                f"[{colorize(self.name, 'yellow')}]: Failed to trigger the site checking cycle"
            ),
            raise_runtime_error=False,
        )

    @log(prefix=_prefix_log_site)
    def discover_services(self) -> None:
        run_command(
            ["sudo", "su", "-", self.name, "-c", "cmk -vI ; cmk -O"],
            error_message=f"[{colorize(self.name, 'yellow')}]: Failed to discover services",
            raise_runtime_error=False,
        )

    @log(prefix=_prefix_log_site, max_level=logging.DEBUG)
    def get_site_connection_config(
        self,
        http_address: str,
        livestatus_port: int,
        message_broker_port: int | None = None,
    ) -> RemoteSiteConnectionConfig:
        internal_url = f"http://{http_address}/{self.name}/check_mk/"
        config: RemoteSiteConnectionConfig = {
            "basic_settings": {
                "alias": f"The  {self.name}",
                "site_id": self.name,
            },
            "status_connection": {
                "connection": {
                    "socket_type": "tcp",
                    "host": http_address,
                    "port": livestatus_port,
                    "encrypted": True,
                    "verify": True,
                },
                "proxy": {
                    "use_livestatus_daemon": "direct",
                },
                "connect_timeout": 2,
                "persistent_connection": False,
                "url_prefix": f"/{self.name}/",
                "status_host": {"status_host_set": "disabled"},
                "disable_in_status_gui": False,
            },
            "configuration_connection": {
                "enable_replication": True,
                "url_of_remote_site": internal_url,
                "disable_remote_configuration": False,
                "ignore_tls_errors": False,
                "direct_login_to_web_gui_allowed": True,
                "user_sync": {"sync_with_ldap_connections": "disabled"},
                "replicate_event_console": True,
                "replicate_extensions": False,
                "is_trusted": False,
            },
        }
        if message_broker_port:
            config["configuration_connection"]["message_broker_port"] = message_broker_port

        if (
            self.cmk_pkg.edition == Edition.OLD_MANAGED
            or self.cmk_pkg.edition == Edition.ULTIMATEMT
        ):
            # required filed for managed edition
            config["basic_settings"]["customer"] = "provider"
        return config

    @log(prefix=_prefix_log_site, max_level=logging.DEBUG)
    def _append_to_file(self, file_path: Path, content: str) -> None:
        try:
            run_command(
                ["sudo", "su", "-", self.name, "-c", f"tee -a {file_path}"],
                input=content,
                error_message=f"Failed to append to {file_path}",
            )
        except UnicodeEncodeError as e:
            raise RuntimeError("Invalid content encoding") from e

    @log(prefix=_prefix_log_site)
    def add_remote_site_certificate(self, remote_site_name: str) -> None:
        cert_path = Path("/omd/sites", remote_site_name, "etc", "ssl", "ca.pem")
        ca_certificates_path = Path(
            "/omd/sites",
            self.name,
            "etc/check_mk/multisite.d/wato/ca-certificates.mk",
        )
        ssl_certificates_path = Path("/omd/sites", self.name, "var/ssl/ca-certificates.crt")

        cert_der = run_command(
            ("sudo", "openssl", "x509", "-inform", "PEM", "-in", str(cert_path), "-outform", "DER"),
            input=str(cert_path).encode(),
            text=False,
        ).stdout
        cert_pem: str = run_command(
            ("openssl", "x509", "-inform", "DER", "-outform", "PEM"), input=cert_der, text=False
        ).stdout.decode("utf-8")

        # Append certificate to SSL trust store
        self._append_to_file(ssl_certificates_path, cert_pem)

        # Format certificate to be inserted as a single line
        cert_escaped = cert_pem.strip().replace("\n", "\\n")
        trust_entry = (
            "trusted_certificate_authorities.setdefault"
            f'("trusted_cas", []).append("{cert_escaped}")\n'
        )

        # Append certificate trust entry to configuration file
        self._append_to_file(ca_certificates_path, trust_entry)

    @log(prefix=_prefix_log_site)
    def register_host_with_agent(self, host_name: str, gui_user: str, gui_pw: str) -> None:
        cmk_agent_ctl_path = shutil.which("cmk-agent-ctl")
        if not cmk_agent_ctl_path:
            raise RuntimeError("cmk-agent-ctl not found. Please install the Checkmk agent.")

        run_command(
            args=[
                "sudo",
                cmk_agent_ctl_path,
                "-v",
                "register",
                "--hostname",
                host_name,
                "--server",
                "127.0.0.1",
                "--site",
                self.name,
                "--user",
                gui_user,
                "--password",
                gui_pw,
                "--trust-cert",
            ],
            error_message=f"Failed to register host {host_name} with cmk-agent-ctl",
        )


def checkmk_agent_needs_installing() -> bool:
    """Check if the Checkmk agent is installed."""
    cmk_agent_ctl_path = shutil.which("cmk-agent-ctl")
    if cmk_agent_ctl_path:
        return True
    # Check if port 6556 is open
    port_6556_open = run_command(["sudo", "ss", "-tuln"]).stdout.find(":6556 ") != -1

    apt_checkmk_installed = (
        run_command(
            ["dpkg-query", "-W", "-f='${Status}'", "check-mk-agent"],
            check=False,  # if no package is found, status code is not 0
        ).stdout.strip()
        == "'install ok installed'"
    )
    return port_6556_open or apt_checkmk_installed


def parse_int(value: str | None) -> int | None:
    """Parse a str to an integer, returning None if the string is empty."""
    if value is None or value.strip() == "":
        return None
    try:
        return int(value)
    except ValueError:
        return None


@log()
def download_and_install_agent(api: "APIClient") -> None:
    """Download and install the Checkmk agent."""
    download_path = Path("/tmp/cmk-agent.deb")
    api.download_agent(download_path)

    run_command(
        ["sudo", "dpkg", "-i", str(download_path)],
        error_message="Failed to install the Checkmk agent",
    )


def configure_tracing(central_site: Site, remote_sites: list[Site]) -> None:
    # has to be called after the sites have been completely set up, but before starting the sites.
    port = parse_int(omd_config_get(central_site.name, "TRACE_RECEIVE_PORT"))
    if port is None:
        raise RuntimeError("Failed to read the TRACE_RECEIVE_PORT for the central site")

    # we assume that central_site and remote_sites share the same config and version

    if (
        central_site.cmk_pkg.base_version < BaseVersion(2, 4, 0)
        or central_site.cmk_pkg.edition == Edition.OLD_SAAS
    ):
        logger.warning(
            "[%s]: Tracing is not supported for version: %s",
            colorize(central_site.name, "yellow"),
            colorize(str(central_site.cmk_pkg.version), "yellow"),
        )
        return

    for remote_site in remote_sites:
        omd_config_set(remote_site.name, "TRACE_SEND_TARGET", f"http://localhost:{port}")
        omd_config_set(remote_site.name, "TRACE_SEND", "on")

    omd_config_set(central_site.name, "TRACE_RECEIVE", "on")
    omd_config_set(central_site.name, "TRACE_SEND", "on")


def read_default_version() -> CMKPackage:
    raw_version = run_command(["omd", "version", "-b"]).stdout.strip()
    package = parse_version(raw_version)
    if isinstance(package, CMKPackage):
        return package
    raise RuntimeError("🤔 Only found partial version with omd. WTF?")


def parse_version(version: str) -> CMKPackage | PartialCMKPackage:
    """Parse the version string into a cmk package."""

    if match := re.match(
        r"^(\d+\.\d+\.\d+)-(\d+[.-]\d+[.-]\d+)(?:\.(\w+))?$",
        version,
    ):
        try:
            return CMKPackage(
                version=VersionWithReleaseDate(
                    base_version=BaseVersion.from_str(match.group(1)),
                    release_date=datetime.strptime(
                        match.group(2).replace("-", "."), "%Y.%m.%d"
                    ).date(),
                ),
                edition=Edition(match.group(3) or "cee"),
            )
        except ValueError as e:
            raise argparse.ArgumentTypeError(e)
    elif match := re.match(r"^(\d+\.\d+\.\d+)(p|b)(\d+).(\w+)$", version):
        try:
            return CMKPackage(
                version=VersionWithPatch(
                    base_version=BaseVersion.from_str(match.group(1)),
                    patch_type="p" if match.group(2) == "p" else "b",
                    patch=int(match.group(3)),
                ),
                edition=Edition(match.group(4)),
            )
        except ValueError as e:
            raise argparse.ArgumentTypeError(e)
    elif match := re.match(r"^(\d+\.\d+\.\d+).(\w+)$", version):
        try:
            return CMKPackage(
                version=VersionWithPatch(
                    base_version=BaseVersion.from_str(match.group(1)),
                    patch_type="p",
                    patch=0,
                ),
                edition=Edition(match.group(2)),
            )
        except ValueError as e:
            raise argparse.ArgumentTypeError(e)
    elif match := re.match(r"^(\d+\.\d)+", version):
        try:
            return PartialCMKPackage(version)
        except ValueError as e:
            raise argparse.ArgumentTypeError(e)
    else:
        raise argparse.ArgumentTypeError(
            f"'{version}' doesn't match expected format"
            " '[<branch>p|b<patch>.<edition>|<branch>-<YYYY.MM.DD>.<edition>]'"
        )


def interactive_select(options: list[str], default: str) -> str:
    print("Available versions:")
    for i, o in enumerate(options):
        maybe_default = " (selected)" if o == default else ""
        print(f"\t{i + 1} {o}{maybe_default}")

    while True:
        user_input = input("Choose an option: ").strip()
        if not user_input:
            return default
        try:
            if (idx := int(user_input)) <= len(options) and idx > 0:
                return options[idx - 1]
        except ValueError:
            pass
        print("Invalid input. Please enter a number of press Enter for the default")


def interactive_version_select(partial_version: PartialCMKPackage) -> CMKPackage:
    raw_versions = run_command(["omd", "versions", "-b"]).stdout.strip().split("\n")
    most_similar = sorted(raw_versions, key=lambda v: partial_version.similarity(v))[-1]
    selected_version = interactive_select(raw_versions, most_similar)
    version = parse_version(selected_version)
    if isinstance(version, CMKPackage):
        return version
    raise RuntimeError("Oh we select another incomplete version from omd output WTF?")


@dataclass(frozen=True)
class Config:
    cmk_pkg: CMKPackage  # You should define this type properly
    name: str
    verbose: bool
    quiet: int
    distributed: int
    language: Language
    force: bool

    @classmethod
    def from_args(cls, args: argparse.Namespace) -> Self:
        version: CMKPackage | PartialCMKPackage = args.version or read_default_version()
        if isinstance(version, PartialCMKPackage):
            version = interactive_version_select(version)
        return cls(
            cmk_pkg=version,
            name=args.name or cls._default_name(version.version),
            verbose=args.verbose,
            quiet=args.quiet,
            distributed=args.distributed,
            language=Language(args.language or "en"),
            force=args.force,
        )

    @staticmethod
    def _default_name(version: VersionWithPatch | VersionWithReleaseDate | BaseVersion) -> str:
        """Find the default site name based on the Checkmk version."""

        match version:
            case VersionWithPatch(base_version=base_version, patch_type=patch_type, patch=patch):
                return f"v{str(base_version).replace('.', '')}{patch_type}{patch}"
            case VersionWithReleaseDate(base_version=base_version):
                return f"v{str(base_version).replace('.', '')}"
            case BaseVersion():
                return f"v{str(version).replace('.', '')}"


class ArgFormatter(argparse.RawTextHelpFormatter):
    pass


def create_parser() -> argparse.ArgumentParser:
    assert __doc__ is not None, "__doc__ must be a non-None string"
    prog, descr = __doc__.split("\n", 1)

    parser = argparse.ArgumentParser(
        prog=prog,
        description=descr,
        formatter_class=ArgFormatter,
    )
    parser.add_argument("--version", action="version", version=__version__)
    return parser


def setup_parser(parser: argparse.ArgumentParser) -> None:
    """Setup the argument parser for the script."""

    parser.add_argument(
        "version",
        type=parse_version,
        nargs="?",
        help="specify the full omd version\n"
        f"(default: {colorize('omd version -b', 'blue')},"
        f" e.g {colorize('2.4.0-2025.04.07.cce', 'blue')}.)",
    )

    parser.add_argument(
        "-n",
        "--name",
        type=str,
        help="set the site name. Defaults to base version\n"
        "(e.g., 2.4.0 -> v240, 2.4.0p1 -> v240p1).",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="increase output verbosity",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="count",
        default=0,
        help="suppress stderr output.\n('-q': suppress info, '-qq': suppress warnings)",
    )
    parser.add_argument(
        "-d",
        "--distributed",
        default=0,
        type=int,
        help="specify the number of distributed sites to set up.",
    )
    parser.add_argument(
        "-l",
        "--language",
        type=Language,
        default=Language.EN,
        choices=Language,
        help="set the language (default: %(default)s).",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="force site setup from scratch, even if the site with same name and\nversion exists",
    )

    parser.add_argument(
        "--omd-configs",
        nargs="*",
        type=parse_config,
        help="list of valid key=value pairs for omd config (e.g. LIVESTATUS_TCP=on)",
        default=[],
    )


def parse_config(config: str) -> tuple[str, str]:
    """
    Parse a key=value option from the command line.
    """
    if "=" not in config:
        raise argparse.ArgumentTypeError(f"Invalid option format: {config}. Expected key=value.")
    key, value = config.split("=", 1)
    return key.strip(), value.strip()


def find_version_by_site_name(site_name: str) -> str | None:
    """
    Get the list of sitenames that are running by a specific version
    """
    path = Path("/omd/sites", site_name, "version")

    if not path.exists() or not path.is_symlink():
        return None

    try:
        target = path.readlink()  # Read the symlink target
        return target.name
    except OSError as e:
        raise RuntimeError(f"ERROR: {e.strerror}") from e


@log()
def ensure_sudo() -> None:
    """It increases the sudo timeout and refreshes it."""
    run_command(["sudo", "-v"], error_message="Failed to acquire sudo privileges.")


def handle_site_creation(site: Site, force: bool, configs: list[tuple[str, str]]) -> None:
    existing_site_version = find_version_by_site_name(site.name)

    if not existing_site_version or existing_site_version != str(site.cmk_pkg) or force:
        site.delete_site()
        site.ensure_auth_works()
        site.create_site()
        site.configure_site(configs)

    else:
        logger.warning(
            "[%s]: Site already exists with version %s",
            colorize(site.name, "blue"),
            colorize(existing_site_version, "green"),
        )
        logger.warning("Use force option to delete the existing site")


def connect_central_to_remote(
    central_site: Site,
    central_api: APIClient,
    remote_site: Site,
) -> None:
    """
    Set up a distributed site.
    """
    central_site.add_remote_site_certificate(remote_site.name)
    livestatusport = parse_int(omd_config_get(remote_site.name, "LIVESTATUS_TCP_PORT"))
    brokerport = parse_int(omd_config_get(remote_site.name, "RABBITMQ_PORT"))

    if livestatusport is None:
        logger.error(
            "[%s]: Failed to read the Livestatus port",
            colorize(remote_site.name, "red"),
        )
        return
    # Broker port is only supported in 2.4.0 and above
    if brokerport is None:
        logger.warning(
            "[%s]: Failed to read the Broker port",
            colorize(remote_site.name, "red"),
        )

    remote_site_config = remote_site.get_site_connection_config(
        "localhost", livestatusport, brokerport
    )

    site_conns = central_api.list_all_site_connections()
    if site_conns and remote_site.name in site_conns:
        logger.warning(
            "[%s]: Site connection %s already exists",
            colorize(central_site.name, "yellow"),
            remote_site.name,
        )
    else:
        central_api.create_site_connection(remote_site_config)

    # to establish connection to the remote site
    central_api.login_to_remote_site(remote_site.name)


def add_user_to_sudoers() -> None:
    # TODO: duplicate code. this is also available as ./bin/omd-setup-site-for-dev
    # we also have to be able to call this as a standalone script to be able to
    # f12 into sites not create with the official tools
    """Add the current user to the sudoers file."""
    username = getpass.getuser()
    sites = [
        site_name
        for site_name in subprocess.check_output(["omd", "sites", "--bare"]).decode().split("\n")
        if site_name
    ]
    sudoers_config = f"{username} ALL = ({','.join(sites)}) NOPASSWD: /bin/bash\n"
    run_command(
        [
            "sudo",
            "EDITOR=tee",
            "visudo",
            "-f",
            "/etc/sudoers.d/omd-setup-site-for-dev",
        ],
        input=sudoers_config,
        error_message="Failed to add user to sudoers",
    )


def format_validate_installation(cmk_pkg: CMKPackage) -> str:
    return f"Validate installation of {cmk_pkg}..."


@log(message_info=format_validate_installation)
def validate_installation(cmk_pkg: CMKPackage) -> None:
    """Validate Checkmk installation with proper error handling."""
    if not (INSTALLATION_PATH / Path(cmk_pkg.omd_version)).exists():
        raise RuntimeError(
            f"Checkmk {cmk_pkg.version} is not installed.\n"
            f"Install it using:\n cmk-dev-install {cmk_pkg.version.iso_format()} "
            f"--edition {cmk_pkg.edition.value}"
        )


def activate_changes(api: APIClient) -> None:
    """Activate changes in the Checkmk site."""
    result = api.activate_changes()
    if not result:
        logger.warning(
            "[%s]: Nothing to activate",
            colorize(api.site_name, "yellow"),
        )


def ensure_host_exists(
    api: APIClient,
    host_name: str,
) -> None:
    """Ensure the host exists; create it if it does not."""
    if (hosts := api.list_all_hosts()) and host_name in hosts:
        logger.warning(
            "[%s]: Host %s already exists",
            colorize(api.site_name, "yellow"),
            host_name,
        )
    else:
        api.create_host(host_name=host_name)


@log(max_level=logging.DEBUG)
def core_logic(args: argparse.Namespace) -> None:
    """Main business logic separated from error handling."""

    ensure_sudo()
    config = Config.from_args(args)
    cmk_pkg = config.cmk_pkg
    site_name = config.name

    validate_installation(cmk_pkg)

    central_site = Site(site_name, cmk_pkg)
    handle_site_creation(central_site, args.force, args.omd_configs)

    # Distributed setup
    remote_sites: list[Site] = []
    for number in range(1, args.distributed + 1):
        remote_site = Site(f"{central_site.name}_r{number}", config.cmk_pkg, is_remote=True)
        handle_site_creation(remote_site, config.force, args.omd_configs)
        remote_sites.append(remote_site)

    if central_site.cmk_pkg.base_version >= BaseVersion(2, 4):
        try:
            configure_tracing(central_site, remote_sites)
        except RuntimeError:
            logger.warning("could not configure tracing")
            pass

    api = APIClient(site_name=site_name)
    central_site.start_site(api)
    try:
        api.set_user_language(config.language.value)
    except CheckmkAPIException as e:
        logger.warning(
            "[%s]: %s",
            colorize(site_name, "yellow"),
            e,
        )

        pass
    ensure_host_exists(api, site_name)
    if not checkmk_agent_needs_installing():
        download_and_install_agent(api)

    for remote_site in remote_sites:
        remote_api = APIClient(site_name=remote_site.name)
        remote_site.start_site(remote_api)
        try:
            remote_api.set_user_language(config.language.value)
        except CheckmkAPIException as e:
            logger.warning(
                "[%s]: %s",
                colorize(remote_site.name, "yellow"),
                e,
            )
            pass
        ensure_host_exists(remote_api, remote_site.name)
        remote_site.register_host_with_agent(
            host_name=remote_site.name, gui_user=GUI_USER, gui_pw=GUI_PW
        )
        activate_changes(remote_api)
        remote_site.trigger_site_checking_cycle()
        remote_site.discover_services()

        connect_central_to_remote(central_site, api, remote_site)

        # create host for the remote site in the central site
        api.create_host(host_name=remote_site.name, logical_site_name=remote_site.name)

    activate_changes(api)

    central_site.register_host_with_agent(
        host_name=central_site.name, gui_user=GUI_USER, gui_pw=GUI_PW
    )

    central_site.trigger_site_checking_cycle()
    central_site.discover_services()

    logger.info(
        "[%s]: Happy hacking: %s ",
        colorize(central_site.name, "green"),
        colorize(f"http://localhost/{central_site.name}/check_mk", "blue"),
    )


def execute(args: argparse.Namespace) -> int:
    try:
        logger.setLevel(max(logging.INFO - ((args.verbose - args.quiet) * 10), logging.DEBUG))

        if logger.level <= logging.DEBUG:
            logger.debug(
                "%s: %s",
                colorize("cmk-dev-site version:", "cyan"),
                colorize(str(__version__), "green"),
            )
        core_logic(args)

    except RuntimeError as e:
        logger.error(str(e))
        return 1

    return 0


def main(sys_argv: list[str] | None = None) -> int:
    with clean_cli_exit():
        parser: argparse.ArgumentParser = create_parser()
        setup_parser(parser)
        args = parser.parse_args(sys_argv or sys.argv[1:])
        return execute(args)



from urllib.parse import parse_qs, urlparse

import fastapi as fapi
import jwt

from cmk_dev_site.saas.constants import TENANT_ID
from cmk_dev_site.saas.oidc_service import (
    AUTHORIZATION_CODES,
    DEFAULT_USERNAME,
    KEY,
    TokenPayload,
    authorize,
    authorize_login,
    tenant_role_mapping,
    token,
)


def test_authorize_shows_plain_login_form() -> None:
    """Verify the GET /authorize endpoint returns the HTML login form."""
    response = authorize(state="abc", redirect_uri="http://example.com/callback")

    assert response.status_code == 200
    body = bytes(response.body).decode()
    assert '<form method="post" action="/authorize">' in body
    assert f'name="username" value="{DEFAULT_USERNAME}"' in body
    assert 'name="role" value="user"' in body
    assert 'name="role" value="admin" checked' in body
    assert '<button type="submit">Login</button>' in body


def test_authorize_login_redirects_with_generated_code() -> None:
    """Verify the POST /authorize endpoint generates a code and redirects."""
    AUTHORIZATION_CODES.clear()

    response = authorize_login(
        state="abc",
        redirect_uri="http://example.com/callback",
        username="alice@example.com",
        role="user",
    )

    assert response.status_code == 303
    location = response.headers["location"]
    parsed = urlparse(location)
    params = parse_qs(parsed.query)
    code = params["code"][0]
    assert params["state"] == ["abc"]
    assert AUTHORIZATION_CODES[code].username == "alice@example.com"
    assert AUTHORIZATION_CODES[code].user_role == "user"


def test_token_and_tenant_role_mapping_use_selected_login() -> None:
    """Verify the full flow: from code generation to tenant role verification."""
    AUTHORIZATION_CODES.clear()

    # 1. Simulate the form submission
    authorize_response = authorize_login(
        state="state-1",
        redirect_uri="http://example.com/callback",
        username="alice@example.com",
        role="user",
    )
    location = authorize_response.headers["location"]
    code = parse_qs(urlparse(location).query)["code"][0]

    # 2. Exchange the code for a token
    token_response = token(client_id="client-1", code=code)
    token_value = token_response.access_token

    # 3. Validate the JWT payload
    payload = TokenPayload.model_validate(
        jwt.decode(token_value, KEY.public, algorithms=["RS256"], options={"verify_aud": False})
    )
    assert payload.email == "alice@example.com"
    assert payload.user_role == "user"
    assert payload.tenant_id == TENANT_ID

    # 4. Verify the tenant API returns the selected role
    tenants_response = tenant_role_mapping(authorization=f"Bearer {token_value}")
    assert tenants_response.model_dump() == {"tenants": {TENANT_ID: {"user_role": "user"}}}
    assert code not in AUTHORIZATION_CODES


def test_token_rejects_unknown_authorization_code() -> None:
    """Verify that invalid codes return a 400 error."""
    AUTHORIZATION_CODES.clear()

    try:
        token(client_id="client-1", code="unknown")
    except fapi.HTTPException as exc:
        assert exc.status_code == 400
        assert exc.detail == "Invalid authorization code"
    else:
        raise AssertionError("expected invalid authorization code to be rejected")



from pathlib import Path
from typing import Final

OIDC_PORT: Final = 8089
HOST: Final = "127.0.0.1"
URL: Final = f"http://{HOST}:{OIDC_PORT}"

CSE_CONFIG_DIR: Final = Path("/etc/cse")
OIDC_CONFIG_PATH: Final = CSE_CONFIG_DIR / "cognito-cmk.json"
ADMIN_PANEL_CONFIG_PATH: Final = CSE_CONFIG_DIR / "admin_panel_url.json"
TENANT_ID: Final = "092fd467-0d2f-4e0a-90b8-4ee6494f7453"  # a valid fixed uuid



import argparse
import logging
from collections.abc import Iterator

from .cmk_dev_install import Edition
from .cmk_dev_install import main as cmk_dev_install
from .cmk_dev_site import main as cmk_dev_site
from .utils.cli import clean_cli_exit
from .utils.log import get_logger
from .version import __version__

logger = get_logger(__name__)


def setup_parser(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "version",
        nargs="?",
        default=None,
    )
    parser.add_argument(
        "edition",
        type=Edition,
        nargs="?",
        choices=Edition,
        default=None,
    )
    parser.add_argument(
        "-d",
        "--distributed",
        metavar="N",
        help="specify the number of distributed sites to set up.",
        type=int,
        default=0,
    )
    parser.add_argument(
        "-n",
        "--name",
        help="specify name of the site to be created",
    )
    parser.add_argument(
        "--dryrun",
        help="only print command, but do not execute",
        action="store_true",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="increase output verbosity",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="count",
        default=0,
        help="decrease output verbosity",
    )


def install_args(version: str | None, edition: str | None, arg_logging_level: str) -> Iterator[str]:
    if version:
        yield str(version)
        if edition:
            yield "--edition"
            yield str(edition)
    if arg_logging_level:
        yield arg_logging_level


def site_args(distributed: int, arg_logging_level: str, name: str | None) -> Iterator[str]:
    if distributed:
        yield "--distributed"
        yield str(distributed)
    if name:
        yield "--name"
        yield name
    yield "-f"
    if arg_logging_level:
        yield arg_logging_level


def core(
    version: str | None,
    edition: str | None,
    *,
    site_distributed: int,
    site_name: str | None,
    dryrun: bool,
    verbose: int,
    quiet: int,
) -> int:
    # we build the commandline and then feed it to the original tools
    # so what we print and execute is exactly the same...
    arg_logging_level = ""
    if verbose or quiet:
        arg_logging_level = f"-{''.join(['v' * verbose, 'q' * quiet])}"

    args_install = list(install_args(version, edition, arg_logging_level))
    args_site = list(site_args(site_distributed, arg_logging_level, site_name))

    print(
        "cmk-dev-install",
        " ".join(args_install),
        "&&",
        "cmk-dev-site",
        " ".join(args_site),
    )

    if not dryrun:
        result = cmk_dev_install(args_install)
        if result != 0:
            return result
        return cmk_dev_site(args_site)
    return 0


def execute(args: argparse.Namespace) -> int:
    try:
        logger.setLevel(max(logging.INFO - ((args.verbose - args.quiet) * 10), logging.DEBUG))
        return core(
            args.version,
            args.edition,
            site_distributed=args.distributed,
            site_name=args.name,
            dryrun=args.dryrun,
            verbose=args.verbose,
            quiet=args.quiet,
        )
    except RuntimeError as e:
        logger.error(str(e))
        return 1
    return 0


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("--version", action="version", version=__version__)
    return parser


def main(sys_argv: list[str] | None = None) -> int:
    with clean_cli_exit():
        parser = create_parser()
        setup_parser(parser)

        args = parser.parse_args()
        return execute(args)



from datetime import date

from cmk_dev_site.cmk_dev_site import Config
from cmk_dev_site.omd import BaseVersion, VersionWithPatch, VersionWithReleaseDate


def test_default_name_base_version():
    version = BaseVersion(2, 3, 0)
    assert Config._default_name(version) == "v230"  # pyright: ignore[reportPrivateUsage]


def test_default_name_version_with_patch():
    version = VersionWithPatch(BaseVersion(2, 3, 0), "p", 1)
    assert Config._default_name(version) == "v230p1"  # pyright: ignore[reportPrivateUsage]


def test_default_name_version_with_release_date():
    version = VersionWithReleaseDate(BaseVersion(2, 3, 0), date(2025, 1, 1))
    assert Config._default_name(version) == "v230"  # pyright: ignore[reportPrivateUsage]



from datetime import datetime
from pathlib import Path
from unittest.mock import MagicMock

import pytest

from cmk_dev_site.cmk_dev_install import (
    CMK_DOWNLOAD_URL,
    TSBUILD_URL,
    find_cached_deb,
    find_last_release,
    parse_version,
)
from cmk_dev_site.omd import (
    BaseVersion,
    CMKPackage,
    Edition,
    VersionWithPatch,
    VersionWithReleaseDate,
)


def test_parse_version():
    assert parse_version("2.3.0") == BaseVersion(2, 3, 0)
    parsed_version = parse_version("2.3.0p1")
    assert isinstance(parsed_version, VersionWithPatch)
    assert parsed_version.base_version == BaseVersion(2, 3, 0)
    assert parsed_version.patch_type == "p"
    assert parsed_version.patch == 1


def _mock_list_versions_with_date_side_effect(cmk_dates: list[str], tst_dates: list[str]):
    def list_versions_with_date_side_effect(
        url: str, _: BaseVersion
    ) -> list[VersionWithReleaseDate]:
        cmk_versions = [
            VersionWithReleaseDate(BaseVersion(2, 5, 0), datetime.strptime(d, "%Y.%m.%d").date())
            for d in cmk_dates
        ]
        tst_versions = [
            VersionWithReleaseDate(BaseVersion(2, 5, 0), datetime.strptime(d, "%Y.%m.%d").date())
            for d in tst_dates
        ]
        if url == CMK_DOWNLOAD_URL:
            return cmk_versions
        if url == TSBUILD_URL:
            return tst_versions
        return []

    return list_versions_with_date_side_effect


@pytest.mark.parametrize(
    "download_urls,cmk_dates,tst_dates,expected_version",
    [
        pytest.param(
            [CMK_DOWNLOAD_URL, TSBUILD_URL],
            ["2025.11.06", "2025.11.07", "2025.11.08", "2025.11.09", "2025.11.10", "2025.11.11"],
            ["2025.11.09", "2025.11.10", "2025.11.11"],
            "2.5.0-2025.11.11",
            id="both_available",
        ),
        pytest.param(
            [CMK_DOWNLOAD_URL],
            ["2025.11.06", "2025.11.07", "2025.11.08", "2025.11.09", "2025.11.10", "2025.11.11"],
            [],
            "2.5.0-2025.11.11",
            id="only_cmk_available",
        ),
        pytest.param(
            [TSBUILD_URL],
            [],
            ["2025.11.09", "2025.11.10", "2025.11.11"],
            "2.5.0-2025.11.11",
            id="only_tst_available_no_cmk",
        ),
        pytest.param(
            [TSBUILD_URL],
            [],
            [],
            RuntimeError,
            id="No package available",
        ),
    ],
)
def test_find_last_release(
    download_urls: list[str],
    cmk_dates: list[str],
    tst_dates: list[str],
    expected_version: str | type[RuntimeError],
):
    mock_file_server = MagicMock()

    mock_file_server.list_versions_with_date.side_effect = (
        _mock_list_versions_with_date_side_effect(cmk_dates, tst_dates)
    )
    mock_file_server.url_exists.return_value = True

    if expected_version is RuntimeError:
        with pytest.raises(RuntimeError):
            find_last_release(
                download_urls=download_urls,
                file_server=mock_file_server,
                base_version=BaseVersion(2, 5, 0),
                edition=Edition.PRO,
                distro_codename="jammy",
            )
    else:
        result = find_last_release(
            download_urls=download_urls,
            file_server=mock_file_server,
            base_version=BaseVersion(2, 5, 0),
            edition=Edition.PRO,
            distro_codename="jammy",
        )

        assert str(result.version) == expected_version


def _make_pkg() -> CMKPackage:
    return CMKPackage(
        version=VersionWithPatch(base_version=BaseVersion(2, 5, 0), patch_type="p", patch=1),
        edition=Edition.PRO,
        distro_codename="jammy",
    )


def test_find_cached_deb_returns_path_when_exists(tmp_path: Path) -> None:
    pkg = _make_pkg()
    deb = tmp_path / pkg.package_name
    deb.touch()
    assert find_cached_deb(pkg, download_dir=tmp_path) == deb


def test_find_cached_deb_returns_none_when_missing(tmp_path: Path) -> None:
    pkg = _make_pkg()
    assert find_cached_deb(pkg, download_dir=tmp_path) is None



"""cmk-dev-install

Download and install a Checkmk version.

If the version is already installed, it will only be downloaded and installed again if
the "-f" flag is used and no site is running on that version. Either way, the version
will be made the default OMD version. This is done to ease the follow-up use of
`cmk-dev-site`, which will by default use the default OMD version to create sites.

Depending on the specified version, the script looks in the following locations:
    https://download.checkmk.com/checkmk
    https://tstbuilds-artifacts.lan.tribe29.com
The script expects credentials to be stored in ~/.cmk-credentials file in the format
user:password (please follow the instructions https://wiki.lan.checkmk.net/x/aYUuBg)
"""

import argparse
import getpass
import hashlib
import http.client
import json
import logging
import re
import shutil
import subprocess
import sys
import time
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime
from html.parser import HTMLParser
from pathlib import Path
from typing import Any, Literal

import requests

from .omd import (
    BaseVersion,
    CMKPackage,
    Edition,
    GitVersion,
    PartialVersion,
    Version,
    VersionWithPatch,
    VersionWithReleaseCandidate,
    VersionWithReleaseDate,
)
from .utils import run_command as run_command_c
from .utils.cli import clean_cli_exit
from .utils.log import colorize, generate_log_decorator, get_logger
from .version import __version__

logger = get_logger(__name__)
log = generate_log_decorator(logger)

CREDENTIALS_FILE = Path("~/.cmk-credentials").expanduser()
CONFIG_PATH = Path("~/.config/jenkins_jobs/jenkins_jobs.ini").expanduser()
INSTALLATION_PATH = Path("/omd/versions")
TSBUILD_URL = "https://tstbuilds-artifacts.lan.tribe29.com"
CMK_DOWNLOAD_URL = "https://download.checkmk.com/checkmk"
DOWNLOAD_DIR = Path("/tmp")


def find_cached_deb(pkg: CMKPackage, download_dir: Path = DOWNLOAD_DIR) -> Path | None:
    """Return the path to a cached .deb file in download_dir if it exists, else None."""
    path = download_dir / pkg.package_name
    return path if path.exists() else None


def run_command(
    args: Sequence[str],
    check: bool = True,
    error_message: str | None = None,
    raise_runtime_error: bool = True,
    text: bool = True,
    silent: bool = False,
    **kwargs: Any,
) -> subprocess.CompletedProcess[Any]:
    return run_command_c(
        args,
        check=check,
        error_message=error_message,
        raise_runtime_error=raise_runtime_error,
        text=text,
        silent=silent,
        logger=logger,  # slightly hackish: pass in the right logger
        **kwargs,
    )


def build_download_url(url: str, pkg: CMKPackage) -> str:
    """
    Generate the download URL for a given package.
    """
    return f"{url}/{pkg.version.download_folder_name}/{pkg.package_name}"


@log()
def remove_package(pkg_name: str, installed_path: Path) -> None:
    """
    Remove a package using apt.
    """
    # making sure the path is relative to /omd/versions/
    if not installed_path.is_relative_to(INSTALLATION_PATH):
        raise RuntimeError(f"ERROR: Removing package failed: {installed_path} is not a valid path")
    run_command(
        [
            "sudo",
            "rm",
            "-rf",
            str(installed_path),
        ],
    )
    # Check if package is installed first
    check_result = run_command(
        [
            "dpkg",
            "-l",
            pkg_name,
        ],
        raise_runtime_error=False,
    )

    # Only attempt removal if package is installed
    if check_result.returncode == 0:
        run_command(
            [
                "sudo",
                "apt-get",
                "purge",
                "-y",
                pkg_name,
            ],
        )


@log()
def install_package(pkg_path: Path) -> None:
    """
    Install a package using apt from the provided file path.
    """
    run_command(
        [
            "sudo",
            "apt-get",
            "install",
            "-y",
            str(pkg_path),
        ],
        error_message="Failed to install package",
    )


def get_user_pass() -> tuple[str, str]:
    """Get the user and password from the credentials file"""

    try:
        user, password = CREDENTIALS_FILE.read_text(encoding="utf-8").strip().split(":", 1)
        return (user, password)
    except Exception as e:
        raise RuntimeError(
            f"ERROR: Credentials file not found or is not in a correct format: {CREDENTIALS_FILE}"
            ", please follow the instructions https://wiki.lan.checkmk.net/x/aYUuBg"
        ) from e


@dataclass
class DistroVersionInfo:
    version_id: str
    version_codename: str


def get_distro_version_info() -> DistroVersionInfo:
    """Get the distribution version ID and codename for Ubuntu."""
    data = dict(
        line.split("=", 1)
        for line in Path("/etc/os-release").read_text(encoding="utf-8").splitlines()
        if "=" in line
    )
    if data.get("ID") == "ubuntu" and {"VERSION_ID", "VERSION_CODENAME"} <= data.keys():
        return DistroVersionInfo(
            version_id=data["VERSION_ID"].strip('"'),
            version_codename=data["VERSION_CODENAME"].strip('"'),
        )
    raise RuntimeError("ERROR: Unsupported distribution or missing version info.")


def parse_version(
    version: str,
) -> Version:
    """Parse the version string into a Version object."""

    if match := re.match(r"^(\d+\.\d+\.\d+)-daily$", version):
        return VersionWithReleaseDate(
            base_version=BaseVersion.from_str(match.group(1)),
            release_date=datetime.today().date(),
        )

    if match := re.match(r"^(\d+\.\d)+$", version):
        return PartialVersion.from_str(match.group(1))

    if match := re.match(r"^(\d+\.\d+\.\d+)$", version):
        return BaseVersion.from_str(match.group(1))

    if match := re.match(r"^(\d+\.\d+(?:\.\d+)?)(p|b)(\d+)$", version):
        return VersionWithPatch(
            base_version=BaseVersion.from_str(match.group(1)),
            patch_type="p" if match.group(2) == "p" else "b",
            patch=int(match.group(3)),
        )

    if match := re.match(r"^(\d+\.\d+(?:\.\d+)?)(p|b)(\d+)-rc(\d+)$", version):
        return VersionWithReleaseCandidate(
            base_version=BaseVersion.from_str(match.group(1)),
            patch_type="p" if match.group(2) == "p" else "b",
            patch=int(match.group(3)),
            rc=int(match.group(4)),
        )
    if match := re.match(r"^git:(.+?):(.+?)$", version):
        return GitVersion(
            branch=match.group(1),
            commit_hash=match.group(2),
        )

    if match := re.match(r"^(\d+\.\d+\.\d+)-(\d+[.-]\d+[.-]\d+)$", version):
        return VersionWithReleaseDate(
            base_version=BaseVersion.from_str(match.group(1)),
            release_date=datetime.strptime(match.group(2).replace(".", "-"), "%Y-%m-%d").date(),
        )

    raise argparse.ArgumentTypeError(
        f"{version!r} doesn't match expected format '2.2.0p23|2.2.0-YYYY-MM-DD|2.2.0-YYYY.MM.DD|"
        "2.2|2.2.0-daily|git:<branch>:<commit_hash>'"
    )


class FileServer:
    """Represents a file server to download packages from."""

    def __init__(
        self,
        user: str,
        password: str,
    ):
        self.user = user
        self.password = password
        self._session = requests.Session()
        self._session.auth = (user, password)
        self._timeout = 120

    def _get(self, url: str, **kwargs: Any) -> requests.Response:
        try:
            response = self._session.get(url, timeout=self._timeout, **kwargs)
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            raise RuntimeError(f"ERROR: {e}") from e

        return response

    @log(max_level=logging.DEBUG)
    def download_hash(self, download_url: str) -> str:
        """Download the hash file for a given file."""
        response = self._get(f"{download_url}.hash")

        hash_data = response.text.strip().split()

        if not hash_data:
            raise ValueError("Empty or invalid hash file.")

        return hash_data[0]

    def _calculate_file_hash(self, file_path: Path, hash_type: str = "sha256") -> str:
        """
        Calculate the hash of a file.

        :param file_path: Path to the file.
        :param hash_type: Hash algorithm (e.g., 'sha256', 'md5', 'sha1').
        :return: Hexadecimal hash of the file.
        """
        try:
            with open(file_path, "rb") as f:
                digest = hashlib.file_digest(f, hash_type)
            return digest.hexdigest()
        except FileNotFoundError as e:
            raise RuntimeError(f"ERROR: File not found: {file_path}") from e

    @log()
    def verify_hash(self, download_url: str, file_path: Path) -> bool:
        """
        Verify the hash of a file.
        :param download_url: The URL of the hash file.
        :param file_path: Path to the file.
        :return: True if the hash matches, False otherwise.
        """
        file_hash = self.download_hash(download_url)
        calculated_hash = self._calculate_file_hash(file_path)

        return file_hash == calculated_hash

    @log()
    def download_package(
        self,
        url: str,
        download_path: Path,
    ) -> None:
        """
        Downloads a package from the file server.

        :param url: The URL of the package.
        :param download_path: The local path where the package will be saved.
        """

        start = time.monotonic()
        response = self._get(
            url,
            stream=True,
        )

        file_size = int(response.headers["Content-Length"])
        with open(download_path, "wb") as file:
            downloaded = 0
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                file.write(chunk)
                downloaded += len(chunk)
                progress = (downloaded / file_size) * 100
                download_rate = downloaded / 1024 / 1024 / (time.monotonic() - start)
                logger.progress(f"Downloading... {progress:.2f}% ({download_rate:.2f} MB/s)")

    def _query_available_versions(self, url: str) -> Sequence[VersionWithReleaseDate]:
        response = self._get(url)
        vs_parser = VersionParser()
        vs_parser.feed(response.text)
        return vs_parser.versions

    @log(max_level=logging.DEBUG)
    def list_versions_with_date(
        self, url: str, base_version: BaseVersion
    ) -> list[VersionWithReleaseDate]:
        return [vs for vs in self._query_available_versions(url) if vs.base_version == base_version]

    @log()
    def query_latest_base_version(self, urls: list[str]) -> BaseVersion:
        # if we have no version *at all*, this raises.
        return max(
            (v.base_version for url in urls for v in self._query_available_versions(url)),
        )

    @log(max_level=logging.DEBUG)
    def url_exists(self, url: str) -> bool:
        response = self._session.head(
            url,
            timeout=self._timeout,
        )
        return response.status_code == 200


class VersionParser(HTMLParser):
    """
    This class is used to parse the versions from the Checkmk download page with re.Pattern.
    """

    def __init__(self) -> None:
        super().__init__()
        self._versions: list[VersionWithReleaseDate] = []
        self._pattern = re.compile(r"^(\d+\.\d+\.\d+)-(\d+.\d+.\d+)")

    @property
    def versions(self) -> Sequence[VersionWithReleaseDate]:
        return self._versions

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
        attrs_dict = dict(attrs)

        if tag == "a" and (href := attrs_dict.get("href")) and (match := self._pattern.match(href)):
            self._versions.append(
                VersionWithReleaseDate(
                    base_version=BaseVersion.from_str(match.group(1)),
                    release_date=datetime.strptime(match.group(2), "%Y.%m.%d").date(),
                )
            )


@log()
def validate_jenkins_jobs_ini() -> bool:
    """
    Validate the jenkins_jobs.ini file exists or not.
    """

    if not CONFIG_PATH.exists():
        this_user_name = run_command(["git", "config", "user.email"]).stdout.strip().split("@")[0]
        config_content = f"""
In order to interact with Jenkins you need to have a file at
"~/.config/jenkins_jobs/jenkins_jobs.ini" containing the following content

# start of file
[jenkins]
user={this_user_name}
# Get the APIKEY from the CI web UI, click top right Profile -> Security -> Add new Token
# https://ci.lan.tribe29.com/user/{this_user_name}/security
password=API_KEY_NOT_YOUR_PASSWORD
url=https://ci.lan.tribe29.com
query_plugins_info=False
# end of file
"""
        logger.error(
            config_content,
        )
        return False
    return True


@dataclass
class ArtifactsResult:
    artifacts: list[str]
    result: str


@log()
def build_install_git_version(
    branch: str, commit_hash: str, edition: Edition, distro_id: str
) -> Path:
    """
    Build and install a version from a specific branch and commit hash.

    :param branch: The branch to build the version from.
    :param commit_hash: The commit hash to build the version from.
    :param edition: The edition of the version.
    :param distro_id: The distro ID to build the version for.
    :return: The path to the built package.
    """

    if not validate_jenkins_jobs_ini():
        raise RuntimeError("ERROR: Jenkins jobs ini file is not found")

    if not shutil.which("ci-artifacts"):
        raise RuntimeError(
            "'ci-artifacts' command not found."
            "To have 'ci-artifacts' available, run 'pip install checkmk-dev-tools' and try again"
        )

    result = run_command(
        [
            "ci-artifacts",
            "fetch",
            f"checkmk/{branch}/builders/trigger-cmk-distro-package",
            "--out-dir",
            "/tmp/",
            "--no-remove-others",
            f"--params=DISTRO=ubuntu-{distro_id},EDITION={edition.name.lower()},CUSTOM_GIT_REF={commit_hash}",
        ],
    )
    typed_result = ArtifactsResult(**json.loads(result.stdout))
    if typed_result.result == "FAILURE":
        raise RuntimeError("ERROR: Failed to build the version.")

    filtered_artifacts = [
        artifact
        for artifact in typed_result.artifacts
        if "check-mk" in artifact and artifact.endswith(".deb")
    ]

    if not filtered_artifacts:
        raise RuntimeError("ERROR: No artifacts found for the version.")

    return Path("/tmp", filtered_artifacts[0])


@log(show_result=True)
def find_last_release(
    download_urls: list[str],
    file_server: FileServer,
    base_version: BaseVersion,
    edition: Edition,
    distro_codename: str,
) -> CMKPackage:
    """
    Find the last release date for a version.
    """
    url_version_date = [
        (url, v_date)
        for url in download_urls
        for v_date in file_server.list_versions_with_date(url, base_version)
    ]
    url_version_date.sort(key=lambda p: p[1].release_date, reverse=True)

    for url, version_date in url_version_date:
        pkg = CMKPackage(
            version=version_date,
            edition=edition,
            distro_codename=distro_codename,
        )
        if file_server.url_exists(build_download_url(url, pkg)):
            return pkg

    raise RuntimeError(f"ERROR: No release found for the version {base_version}")


@log(max_level=logging.DEBUG)
def find_sitenames_by_version(version: str) -> list[str]:
    """
    Get the list of sitenames that are running by a specific version
    """
    site_names: list[str] = []
    try:
        for path in Path("/omd/sites").glob("*"):
            path = path / "version"
            if path.is_symlink():
                target = path.readlink()  # Read the symlink target
                site_version = target.name  # Extract the version from the target path
                site_name = path.parent.name
                if site_version == version:
                    site_names.append(site_name)
        return site_names
    except OSError as e:
        raise RuntimeError(f"ERROR: {e.strerror}") from e


def apply_acls_to_version(version: str) -> None:
    """
    Apply ACLs to omd version.
    """
    version_path = INSTALLATION_PATH / version
    user = getpass.getuser()

    if not version_path.exists():
        raise RuntimeError(f"ERROR: Version {version_path.name} does not exist")
    run_command(
        [
            "sudo",
            "setfacl",
            "--recursive",
            "-m",
            f"user:{user}:rwX",
            str(version_path),
        ],
    )


def get_default_version() -> str:
    return run_command(["sudo", "omd", "version", "-b"]).stdout.strip()


@log(max_level=logging.DEBUG)
def set_default_version(version: str) -> None:
    """
    Set the default version to the specified version.
    """
    if version == get_default_version():
        return
    run_command(
        ["sudo", "omd", "setversion", version],
    )


@log()
def ensure_sudo() -> None:
    """It increases the sudo timeout and refreshes it."""
    run_command(["sudo", "-v"])


class ArgFormatter(argparse.RawTextHelpFormatter):
    pass


class DevInstallArgs(argparse.Namespace):
    build: Version
    edition: Edition
    force: bool
    verbose: int
    quiet: int
    download_only: bool


def create_parser() -> argparse.ArgumentParser:
    assert __doc__ is not None, "__doc__ must be a non-None string"
    prog, descr = __doc__.split("\n", 1)

    parser = argparse.ArgumentParser(
        prog=prog,
        description=descr,
        formatter_class=ArgFormatter,
    )
    parser.add_argument("--version", action="version", version=__version__)
    return parser


def setup_parser(parser: argparse.ArgumentParser) -> None:
    """Setup the argument parser for the script."""

    parser.add_argument(
        "build",
        metavar="version",
        type=parse_version,
        nargs="?",
        help=f"""specify the version in one of the following formats:

  {colorize("2.4.0-daily", "green")}: install today's daily version
  {colorize("2.4", "green")}: install the latest available daily build
  {colorize("2.4.0-2025-01-01", "green")}: install specific daily version
  {colorize("2.4.0p23", "green")}: install released patch version
  {colorize("git:master:39f57c98f92", "green")}: Build and install from a specific
    branch and commit

Per default it will try to install the daily version of the
latest branch that can be found.
""",
    )

    parser.add_argument(
        "-e",
        "--edition",
        type=Edition,
        default=None,
        choices=Edition,
        help="specify the edition of the version to install (default: %(default)s).",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="force re-downloading and installing the version.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="increase output verbosity",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="count",
        default=0,
        help="suppress stderr output.'-q': suppress info, '-qq': suppress warnings.",
    )
    parser.add_argument(
        "-d",
        "--download-only",
        action="store_true",
        help="download only (do not install)",
    )


@log(max_level=logging.DEBUG)
def validate_installation(cmk_pkg: CMKPackage, force: bool, download_only: bool) -> bool:
    """
    Validate if the installation should go through.
    """
    already_installed = (INSTALLATION_PATH / Path(cmk_pkg.omd_version)).exists()

    if already_installed and not (force or download_only):
        logger.warning(
            "Version %s already installed.",
            INSTALLATION_PATH / Path(cmk_pkg.omd_version),
        )
        logger.warning("Use --force to re-download and re-install.")

        return False

    if (
        already_installed
        and (existed_sitenames := find_sitenames_by_version(cmk_pkg.omd_version))
        and not download_only
    ):
        nl = "\n"
        raise RuntimeError(
            f"Found sites existed with the version {cmk_pkg.omd_version}: {existed_sitenames}\n"
            "Please remove the site, before reinstalling using:\n"
            f"""{
                nl.join(
                    [
                        f"sudo omd -f rm --kill --apache-reload {sitename}"
                        for sitename in existed_sitenames
                    ]
                )
            }"""
        )

    return True


def _fetch_and_verify(
    file_server: FileServer,
    urls: list[str],
    cmk_pkg: CMKPackage,
    download_path: Path,
) -> None:
    """Find a valid download URL, fetch the package, and verify its hash."""
    valid_url = None
    for url in urls:
        logger.debug(f"Checking URL: {url}")
        pkg_url = build_download_url(url, cmk_pkg)

        if not file_server.url_exists(pkg_url):
            logger.warning(f"Version {cmk_pkg} not found in the download server {url}")
        else:
            valid_url = pkg_url
            break

    if valid_url is None:
        raise RuntimeError("Download URL not found.")
    file_server.download_package(
        url=valid_url,
        download_path=download_path,
    )
    if not (file_server.verify_hash(valid_url, download_path)):
        raise RuntimeError("ERROR: Hash verification failed.")


def download_and_install_cmk_pkg(
    file_server: FileServer,
    urls: list[str],
    cmk_pkg: CMKPackage,
    force: bool,
    download_only: bool,
) -> CMKPackage:
    """
    Download and install a Checkmk package.
    """
    if not validate_installation(cmk_pkg, force, download_only):
        set_default_version(cmk_pkg.omd_version)
        return cmk_pkg

    download_path = DOWNLOAD_DIR / cmk_pkg.package_name
    cached = find_cached_deb(cmk_pkg, DOWNLOAD_DIR)
    if cached is not None and not force:
        logger.info("Found cached .deb at %s, skipping download", cached)
    else:
        if cached is not None and force:
            logger.info("--force specified, ignoring cached .deb")
        _fetch_and_verify(file_server, urls, cmk_pkg, download_path)
    if not download_only:
        remove_package(cmk_pkg.package_raw_name, INSTALLATION_PATH / Path(cmk_pkg.omd_version))
        install_package(download_path)
    return cmk_pkg


@log(max_level=logging.DEBUG)
def core_logic(
    version: Version | None, edition: Edition, force: bool, download_only: bool
) -> tuple[str, Path]:
    """
    Download and Install a Checkmk version.

    :param version: The version to install.
    :param edition: The edition of the version.
    :param force: Re-download and re-install the version.
    :param download: Download only (no install).
    :return: The installed version and the package path.
    """
    if not download_only:
        ensure_sudo()

    download_urls = [CMK_DOWNLOAD_URL]
    if edition is Edition.OLD_SAAS or edition is Edition.CLOUD:
        # cloud editions are only available via tstbuilds
        download_urls = [TSBUILD_URL]

    user, password = get_user_pass()
    file_server = FileServer(user=user, password=password)
    distro = get_distro_version_info()

    match version:
        case GitVersion(branch=branch, commit_hash=commit_hash):
            pkg_path = build_install_git_version(branch, commit_hash, edition, distro.version_id)
            install_package(pkg_path)
            # default version will point to the latest installed version
            # we don't have to figure out which git version we just installed
            installed_version = get_default_version()

        case PartialVersion():
            cmk_pkg = find_last_release(
                download_urls, file_server, version, edition, distro.version_codename
            )
            cmk_pkg = download_and_install_cmk_pkg(
                file_server=file_server,
                urls=download_urls,
                cmk_pkg=cmk_pkg,
                force=force,
                download_only=download_only,
            )
            pkg_path = DOWNLOAD_DIR / cmk_pkg.package_name
            installed_version = cmk_pkg.omd_version
        case _:
            if version is None:
                version = VersionWithReleaseDate(
                    base_version=file_server.query_latest_base_version(
                        [f"{u}/" for u in download_urls]
                    ),
                    release_date=datetime.today().date(),
                )
            cmk_pkg = CMKPackage(
                version=version,
                edition=edition,
                distro_codename=distro.version_codename,
            )
            cmk_pkg = download_and_install_cmk_pkg(
                file_server=file_server,
                urls=download_urls,
                cmk_pkg=cmk_pkg,
                force=force,
                download_only=download_only,
            )
            pkg_path = DOWNLOAD_DIR / cmk_pkg.package_name
            installed_version = cmk_pkg.omd_version

    if not download_only:
        apply_acls_to_version(installed_version)

    return installed_version, pkg_path


def validate_version_edition(
    version: Version | None, edition: Edition | None
) -> tuple[Version | None, Edition]:
    # editions changed with the release of 2.5.0
    # we want to find out which editions are valid with the current version
    # if version is None, we install the newest version, so we are clearly post240:
    editions_to_use: Literal["pre250", "post240", "dontknow"] = "post240"
    if version is not None:
        # TODO: Need refactoring
        if isinstance(
            version,
            (VersionWithReleaseDate, VersionWithPatch, VersionWithReleaseCandidate),
        ):
            base_version = version.base_version
        else:
            base_version = version
        match base_version:
            case GitVersion():
                editions_to_use = "dontknow"
            case _:
                if base_version.major != 2:
                    raise RuntimeError("the future is now!")
                if base_version.minor < 5:
                    editions_to_use = "pre250"

    # first we want to figure out the default edition,
    # if no explicit edition is specified by the user:
    if edition is None:
        if editions_to_use == "pre250":
            edition = Edition.OLD_ENTERPRISE
        elif editions_to_use == "post240":
            edition = Edition.PRO
        else:
            raise RuntimeError("Please specify edition if you want to build from git")

    # the user can mix editions and version, but we want to make sure
    # only versions and editions can continue that actually match:
    valid_post_240_editions = {
        Edition.COMMUNITY,
        Edition.PRO,
        Edition.ULTIMATEMT,
        Edition.ULTIMATE,
        Edition.CLOUD,
    }
    if editions_to_use == "post240" and edition not in valid_post_240_editions:
        raise RuntimeError(
            "You want to install version 2.5.0 or higher, "
            "you have to use the new Edition specifier."
            f"\nGot {edition}, expected {', '.join(sorted(valid_post_240_editions))}"
        )

    valid_pre_250_editions = {
        Edition.OLD_RAW,
        Edition.OLD_ENTERPRISE,
        Edition.OLD_MANAGED,
        Edition.OLD_CLOUD,
        Edition.OLD_SAAS,
    }
    if editions_to_use == "pre250" and edition not in valid_pre_250_editions:
        raise RuntimeError(
            "You want to install version 2.4.0 or smaller, "
            "you have to use the old Edition specifier."
            f"\nGot {edition}, expected {', '.join(sorted(valid_pre_250_editions))}"
        )

    return version, edition


def execute(args: argparse.Namespace) -> int:
    logger.setLevel(max(logging.INFO - ((args.verbose - args.quiet) * 10), logging.DEBUG))

    if args.verbose - args.quiet > 3:
        # disable the custom logging, otherwise we see all messages twice
        logger.removeHandler(logger.handlers[0])
        logging.basicConfig(level=logging.DEBUG)
        logger_urllib = logging.getLogger("requests.packages.urllib3")
        logger_urllib.setLevel(logging.DEBUG)
        logger_urllib.propagate = True

    if args.verbose - args.quiet > 4:
        http.client.HTTPConnection.debuglevel = 1

    if logger.level <= logging.DEBUG:
        logger.debug(
            "%s: %s",
            colorize("cmk-dev-site version:", "cyan"),
            colorize(str(__version__), "green"),
        )
    try:
        version, edition = validate_version_edition(args.build, args.edition)
        installed_version, pkg_path = core_logic(version, edition, args.force, args.download_only)
    except RuntimeError as e:
        logger.error(e)
        return 1

    print(installed_version)
    if args.download_only:
        print(pkg_path)
    return 0


def main(sys_argv: list[str] | None = None) -> int:
    with clean_cli_exit():
        parser = create_parser()
        setup_parser(parser)

        args = parser.parse_args(sys_argv or sys.argv[1:], namespace=DevInstallArgs())
        return execute(args)



import time
from pathlib import Path
from typing import (
    Any,
    NotRequired,
    TypedDict,
    TypeVar,
)

import requests
from requests.exceptions import JSONDecodeError

T = TypeVar("T")


class BasicSettings(TypedDict):
    alias: str
    site_id: str
    customer: NotRequired[str]


class ConnectionDetails(TypedDict):
    socket_type: str
    host: str
    port: int
    encrypted: bool
    verify: bool


class StatusHost(TypedDict):
    status_host_set: str


class StatusConnection(TypedDict):
    connection: ConnectionDetails
    proxy: dict[str, str]
    connect_timeout: int
    persistent_connection: bool
    url_prefix: str
    status_host: StatusHost
    disable_in_status_gui: bool


class UserSync(TypedDict):
    sync_with_ldap_connections: str


class ConfigurationConnection(TypedDict):
    enable_replication: bool
    url_of_remote_site: str
    disable_remote_configuration: bool
    ignore_tls_errors: bool
    direct_login_to_web_gui_allowed: bool
    user_sync: UserSync
    replicate_event_console: bool
    replicate_extensions: bool
    message_broker_port: NotRequired[int]
    is_trusted: bool


class RemoteSiteConnectionConfig(TypedDict):
    basic_settings: BasicSettings
    status_connection: StatusConnection
    configuration_connection: ConfigurationConnection


class CheckmkAPIException(Exception):
    """Base exception for Checkmk API errors."""

    def __init__(
        self,
        message: str,
        status_code: int | None = None,
        response_data: dict[Any, Any] | str | None = None,
    ):
        super().__init__(message)
        self.status_code = status_code
        self.response_data = response_data

    def __str__(self) -> str:
        parts = [super().__str__()]
        if self.status_code is not None:
            parts.append(f"Status code: {self.status_code}")
        if self.response_data is not None:
            parts.append(f"Response: {self.response_data}")
        return "\n".join(parts)


def build_exception(response: requests.Response, message: str) -> CheckmkAPIException:
    try:
        return CheckmkAPIException(
            message=message,
            status_code=response.status_code,
            response_data=response.json(),
        )
    except JSONDecodeError:
        return CheckmkAPIException(
            message=message + "\nFailed to parse response",
            status_code=response.status_code,
            response_data=response.text if response.text else None,
        )


class APIClient:
    """Checkmk API client."""

    def __init__(
        self,
        site_name: str = "heute",
        username: str = "cmkadmin",
        password: str = "cmk",
        # Note: don't be confused with the hosts registered on checkmk
        server_host_name: str = "localhost",
        proto: str = "http",
    ) -> None:
        self.base_url = f"{proto}://{server_host_name}/{site_name}/check_mk/api/1.0"

        self.site_name = site_name
        self.session = requests.session()
        self.session.headers["Authorization"] = f"Bearer {username} {password}"
        self.session.headers["Accept"] = "application/json"

    def version(self) -> dict[str, str] | None:
        response = self.session.get(
            f"{self.base_url}/version",
            headers={
                "Content-Type": "application/json",
            },
        )
        if response.status_code == 200:
            return response.json()
        else:
            return None

    def create_host(
        self,
        host_name: str,
        logical_site_name: str | None = None,
        ip_address: str = "127.0.0.1",
    ) -> None:
        """Create a host on the Checkmk server."""
        logical_site_name = logical_site_name or self.site_name
        response = self.session.post(
            f"{self.base_url}/domain-types/host_config/collections/all",
            params={
                "bake_agent": False,
            },
            headers={
                "Content-Type": "application/json",
            },
            json={
                "host_name": host_name,
                "folder": "/",
                "attributes": {
                    "ipaddress": ip_address,
                    "site": logical_site_name,
                },
            },
        )
        if response.status_code != 200:
            raise build_exception(response, message="Failed to create host")

    def list_all_hosts(self) -> list[str]:
        """List all hosts registered on the Checkmk server."""
        response = self.session.get(
            f"{self.base_url}/domain-types/host_config/collections/all",
            headers={
                "Content-Type": "application/json",
            },
        )
        if response.status_code != 200:
            raise build_exception(response, message="Failed to list all hosts")

        hosts = [host["title"] for host in response.json()["value"]]
        return hosts

    def _get_href_from_links(self, links: list[dict[str, str]], name: str) -> str | None:
        """Extract the href from the links."""
        for link in links:
            if link["rel"] == name:
                return link["href"]
        return None

    def _get(self, url: str) -> requests.Response:
        """Make a GET request to the Checkmk server."""
        full_url = url if url.startswith(("http://", "https://")) else f"{self.base_url}{url}"
        return self.session.get(full_url)

    def create_site_connection(self, site_config: RemoteSiteConnectionConfig) -> None:
        """Create a site connection on the Checkmk server."""
        response = self.session.post(
            f"{self.base_url}/domain-types/site_connection/collections/all",
            headers={
                "Content-Type": "application/json",
            },
            json={"site_config": site_config},
        )

        if response.status_code != 200:
            raise build_exception(response, "Failed to create site connection")

    def set_user_language(self, language: str) -> None:
        """Set the user language on the Checkmk server."""
        response = self.session.put(
            f"{self.base_url}/objects/user_config/cmkadmin",
            headers={
                # (required) The value of the, to be modified, object's ETag header.
                "If-Match": "*",
                "Content-Type": "application/json",
            },
            json={"language": language},
        )
        if response.status_code != 200:
            raise build_exception(response, "Failed to set user language")

    def list_all_site_connections(self) -> list[str]:
        """List all site connections registered on the Checkmk server."""
        response = self.session.get(
            f"{self.base_url}/domain-types/site_connection/collections/all",
            headers={
                "Content-Type": "application/json",
            },
        )
        if response.status_code != 200:
            raise build_exception(response, "Failed to list all site connections")

        site_connections = [site["id"] for site in response.json()["value"]]
        return site_connections

    def activate_changes(self) -> list[str] | None:
        """Activate pending changes on the Checkmk server.

        Returns:
            list: A list of activated changes if any were made.
            None: If there were no changes to activate.
        """
        response = self.session.post(
            f"{self.base_url}/domain-types/activation_run/actions/activate-changes/invoke",
            headers={
                # (required) The value of the, to be modified, object's ETag header.
                "If-Match": "*",
                # (required) A header specifying which type of content
                # is in the request/response body.
                "Content-Type": "application/json",
            },
            json={
                "redirect": False,
                "sites": [],
                "force_foreign_changes": True,
            },
            allow_redirects=True,
        )

        if response.status_code == 422:
            return None
        elif response.status_code != 200:
            raise build_exception(response, "Failed to activate changes")

        changes: set[str] = set()
        # Polling loop
        link = self._get_href_from_links(response.json()["links"], "self")
        if link is None:
            raise build_exception(
                response,
                "Activate change : Failed to find 'self' link in the response",
            )

        while True:
            # Fetch data
            response = self._get(link)
            if response.status_code != 200:
                raise build_exception(response, "Failed to fetch activation changes")
            data = response.json()

            # Process changes
            for change in data["extensions"]["changes"]:
                change_id = change["id"]
                if change_id not in changes:
                    changes.add(change_id)

            # Check if processing is complete
            if not data["extensions"]["is_running"]:
                return list(changes)
            # Wait before polling again
            time.sleep(1)

    def login_to_remote_site(
        self, site_id: str, user: str = "cmkadmin", password: str = "cmk"
    ) -> None:
        response = self.session.post(
            f"{self.base_url}/objects/site_connection/{site_id}/actions/login/invoke",
            headers={
                "Content-Type": "application/json",
            },
            json={"username": user, "password": password},
        )

        if response.status_code != 204:
            raise build_exception(response, "Failed to login to remote site")

    def download_agent(self, download_path: Path) -> None:
        """Download the Checkmk agent."""
        response = self.session.get(
            f"{self.base_url}/domain-types/agent/actions/download/invoke",
            params={"os_type": "linux_deb"},
            headers={
                "Accept": "application/octet-stream",
                "Content-Type": "application/json",
            },
        )
        if response.status_code == 200:
            with open(download_path, "wb") as f:
                f.write(response.content)
        else:
            raise build_exception(response, "Failed to download agent")
